from functools import partial
from ...utilities import default_for_serialisation
import rapidjson as json
dumper = partial(json.dumps, default=default_for_serialisation)
from ...parsers.blast_utils import BlastOpener
from .xml_utils import _get_query_for_blast, _get_target_for_blast
from xml.parsers.expat import ExpatError
import xml
from ...utilities.dbutils import connect
from sqlalchemy.orm.session import Session
import msgpack
import tempfile
import os
from .query import Query
from .target import Target
from .hit import prepare_hit, InvalidHit
from .xml_utils import get_multipliers, get_off_by_one
from .utils import load_into_db
import multiprocessing as mp
from ...utilities.log_utils import create_null_logger
import pandas as pd
import zlib
import struct


struct_row = struct.Struct(">LLL")


def xml_pickler(configuration, filename, default_header,
                cache=None,
                max_target_seqs=10):
    valid, _, exc = BlastOpener(filename).sniff(default_header=default_header)
    engine = connect(configuration)
    session = Session(bind=engine)

    if not valid:
        err = "Invalid BLAST file: %s" % filename
        raise TypeError(err)
    directory = os.path.dirname(filename)
    with tempfile.NamedTemporaryFile(suffix=".db", dir=directory, mode="wb", delete=False) as dbhandle:
        dbname = dbhandle.name
        if not isinstance(cache, dict) or set(cache.keys()) != {"query", "target"}:
            cache = dict()
            cache["query"] = dict((item.query_name, item.query_id) for item in session.query(Query))
            cache["target"] = dict((item.target_name, item.target_id) for item in session.query(Target))

        rows = b""
        try:
            with BlastOpener(filename) as opened:
                try:
                    qmult, tmult = None, None
                    for query_counter, record in enumerate(opened, start=1):
                        if qmult is None:
                            qmult, tmult = get_multipliers(record)
                            off_by_one = get_off_by_one(record)
                        hits, hsps, cache = objectify_record(
                            record, [], [], cache,
                            qmult=qmult, tmult=tmult, off_by_one=off_by_one)

                        record = {"hits": hits, "hsps": hsps}
                        jrecord = dumper(record)
                        write_start = dbhandle.tell()
                        write_length = dbhandle.write(msgpack.dumps(jrecord))
                        rows += struct_row.pack(query_counter, write_start, write_length)

                except ExpatError as err:
                    raise ExpatError("{} is an invalid BLAST file, sending back anything salvageable.\n{}".format(filename,
                                                                                                                  err))
        except xml.etree.ElementTree.ParseError as err:
            # logger.error("%s is an invalid BLAST file, sending back anything salvageable", filename)
            raise xml.etree.ElementTree.ParseError(
                "{} is an invalid BLAST file, sending back anything salvageable.\n{}".format(filename, err))
        except ValueError as err:
            # logger.error("Invalid BLAST entry")
            raise ValueError(
                "{} is an invalid BLAST file, sending back anything salvageable.\n{}".format(filename, err))

        return dbname, zlib.compress(rows)


def _serialise_xmls(self):
    # Load sequences in DB, precache IDs

    if isinstance(self.xml, str):
        self.xml = [self.xml]
    else:
        assert isinstance(self.xml, (list, set))

    hits, hsps = [], []
    hit_counter, record_counter = 0, 0

    for filename in self.xml:
        valid, header, exc = BlastOpener(filename).sniff()
        if valid is True:
            self.header = header
        else:
            self.logger.error(exc)
            self.xml.remove(filename)

    cache = {"query": self.queries, "target": self.targets}
    if self._blast_loading_debug is False and (self.single_thread is True or self.procs == 1):
        for filename in self.xml:
            valid, _, exc = BlastOpener(filename).sniff(default_header=self.header)
            if not valid:
                self.logger.error(exc)
                continue
            try:
                self.logger.debug("Analysing %s", filename)
                qmult, tmult = None, None
                with BlastOpener(filename) as opened:
                    for record in opened:
                        if qmult is None:
                            qmult, tmult = get_multipliers(record)
                            off_by_one = get_off_by_one(record)
                        record_counter += 1
                        if record_counter > 0 and record_counter % 10000 == 0:
                            self.logger.info("Parsed %d queries", record_counter)
                        current = len(hits)
                        hits, hsps, cache = objectify_record(
                            record, hits, hsps,
                            qmult=qmult,
                            tmult=tmult,
                            cache=cache,
                            logger=self.logger, off_by_one=off_by_one)
                        hit_counter += len(hits) - current
                        hits, hsps = load_into_db(self, hits, hsps, force=False)
                self.logger.debug("Finished %s", filename)
            except ExpatError:
                self.logger.error("%s is an invalid BLAST file, saving what's available", filename)
        _, _ = load_into_db(self, hits, hsps, force=True)
    elif self._blast_loading_debug is True or self.procs > 1:
        self.logger.warning("Creating a pool with %d processes", min(self.procs, len(self.xml)))
        results = []
        if self._blast_loading_debug is True:
            for num, filename in enumerate(self.xml):
                results.append(xml_pickler(self.configuration,
                                           filename, self.header,
                                           cache=None,
                                           max_target_seqs=self._max_target_seqs))
        else:
            pool = mp.Pool(self.procs)
            for num, filename in enumerate(self.xml):
                args = (self.configuration, filename, self.header)
                kwds = {"max_target_seqs": self._max_target_seqs, "cache": None}
                pool.apply_async(xml_pickler, args=args, kwds=kwds, callback=results.append)
            pool.close()
            pool.join()

        for dbfile, rows in results:
            with open(dbfile, "rb") as dbhandle:
                for query_counter, write_start, write_length in struct_row.iter_unpack(zlib.decompress(rows)):
                    dbhandle.seek(write_start)
                    result = json.loads(msgpack.loads(dbhandle.read(write_length), raw=False, use_list=True))
                    __hits = result["hits"]
                    __hsps = result["hsps"]
                    assert (not __hsps and not __hits) or __hsps != __hits, (__hits, __hsps)
                    hit_counter += len(__hits)
                    hits.extend(__hits)
                    hsps.extend(__hsps)
                    hits, hsps = load_into_db(self, hits, hsps, force=False)
                    if record_counter > 0 and record_counter % 10000 == 0:
                        self.logger.debug("Parsed %d queries", record_counter)
            os.remove(dbfile)

        self.logger.debug("Finished sending off the data for serialisation")
        _, _ = load_into_db(self, hits, hsps, force=True)

    self.logger.info("Loaded %d alignments for %d queries",
                     hit_counter, record_counter)

    self.logger.info("Finished loading blast hits")


def objectify_record(record, hits, hsps, cache,
                     logger=create_null_logger(),
                     qmult=1, tmult=1, off_by_one=False):
    """
    Private method to serialise a single record into the DB.

    :param record: The BLAST record to load into the DB.
    :type record: Bio.SearchIO.QueryResult
    :param hits: Cache of hits to load into the DB.
    :type hits: list

    :param hsps: Cache of hsps to load into the DB.
    :type hsps: list

    :param cache: cache of match query_name/query_id
    :type cache: dict

    :param logger: logger
    :type logger: logging.Logger

    :param qmult: query multiplier (3 for BLASTX or TBLASTX, 1 otherwise)
    :type qmult: int

    :param qmult: query multiplier (1 for TBLASTX or TBLASTN, 1 otherwise)
    :type qmult: int

    :param off_by_one: DIAMOND before 0.9.31 has a bug which causes the query end to be off by 1.
    :type off_by_one: bool

    :returns: hits, hsps, cache
    :rtype: (list, list, dict)
    """

    if len(record.hits) == 0:
        return hits, hsps, cache

    current_query, name, cache["query"] = _get_query_for_blast(record, cache["query"])
    alignments = dict((ccc, (hit.id, max(_.bitscore for _ in hit.hsps))) for ccc, hit in enumerate(record.hits))
    ids, scores = list(zip(*alignments.values()))
    phits = pd.DataFrame({"idx": list(alignments.keys()), "target_name": ids, "bitscore": scores})
    for hit_num, t in enumerate(phits.sort_values(["bitscore", "target_name"],
                                                  ascending=[False, True]).itertuples(name=None)):
        idx, target_name, bitscore = t[1:]
        alignment = record.hits[idx]
        logger.debug("Started the hit %s vs. %s", name, record.hits[idx].id)
        current_target, cache["target"] = _get_target_for_blast(alignment, cache["target"])
        hit_dict_params = dict()
        (hit_dict_params["query_multiplier"],
         hit_dict_params["target_multiplier"]) = (qmult, tmult)
        hit_evalue = min(_.evalue for _ in record.hits[idx].hsps)
        hit_bs = bitscore
        hit_dict_params["hit_number"] = hit_num + 1
        hit_dict_params["evalue"] = hit_evalue
        hit_dict_params["bits"] = hit_bs
        try:
            hit, hit_hsps = prepare_hit(alignment, current_query,
                                        current_target,
                                        query_length=record.seq_len,
                                        off_by_one=off_by_one,
                                        **hit_dict_params)
        except InvalidHit as exc:
            logger.error(exc)
            continue
        hit["query_aligned_length"] = min(record.seq_len, hit["query_aligned_length"])
        hits.append(hit)
        hsps.extend(hit_hsps)

    return hits, hsps, cache



import multiprocessing
from ..parsers import parser_factory
from ..parsers.bam_parser import BamParser
from ..utilities.log_utils import create_queue_logger
from ..utilities import overlap
import logging
import logging.handlers
from .. import exceptions
from sys import intern
import rapidjson as json
import msgpack
import os
from ..transcripts import Transcript
from operator import itemgetter
import random
import struct
import zlib


__author__ = 'Luca Venturini'

# TODO: found_ids can be refactored out, in preference of a unique ID per file which is used as part of the label
#  in doing this, we prevent requiring the user to rename their inputs as there would be no repeated naming.

def __raise_redundant(row_id, name, label):

    if label == '':

        raise exceptions.RedundantNames(
            """{0} has already been found in another file but is present in {1};
               this will cause unsolvable collisions. Please rerun preparation using
               labels to tag each file.""".format(
                row_id, name
            ))
    else:
        raise exceptions.RedundantNames(
            """"{0} has already been found in another file but is present in {1};
            this will cause unsolvable collisions. This happened even if
            you specified label {2}; please change them in order to ensure that
            no collisions happen.""".format(row_id, name, label))


def __raise_invalid(row_id, name, label):

    raise exceptions.InvalidAssembly(
        """{0} is present multiple times in {1}{2}. This breaks the input parsing.
        Please ensure that the file is properly formatted, with unique IDs.""".format(
            row_id,
            name,
            "(label: {0})".format(label) if label != '' else ""))


def _create_split_tobject(tobject: dict, start, stop, num: int):
    """
    Function to create a subset of the transcript, by keeping only the relevant exons
    :param tobject: dictionary of the features
    :param segments: the segments
    :param tid: original name
    :param num: progressive numbering of the transcript
    :return:
    """

    newobj = tobject.copy()
    foundany = False
    for key in tobject:
        if key == "features":
            newobj["features"] = dict.fromkeys(tobject["features"])
            for feature in newobj["features"]:
                newobj["features"][feature] = []
                for ff in tobject["features"][feature]:
                    if overlap((start, stop), (ff[0], ff[1]), positive=True, flank=0) > 0:
                        foundany = True
                        newobj["features"][feature].append(ff[:])
        else:
            newobj[key] = tobject[key]

    newobj["tid"] = newobj["tid"] + f"_isplit.{num}"
    newobj["parent"] = "{}.gene".format(newobj["tid"])
    return newobj, newobj["tid"], start, stop


def _evaluate_tid(tid, tobject, logger, min_length, max_intron):
    if "exon" in tobject["features"]:
        segments = tobject["features"]["exon"][:]
    elif "CDS" in tobject["features"]:
        segments = tobject["features"]["CDS"][:]
        for feature in tobject["features"]:
            if "utr" in feature.lower():
                segments.extend(tobject["features"][feature])
            else:
                continue
        segments = sorted(segments, key=itemgetter(0))
        # Now check the exons
        exons = []
        if len(segments) == 0:
            logger.warning("No valid exon feature for %s, continuing", tid)
            return []
        elif len(segments) == 1:
            exons = segments[0]
        else:
            current = segments[0]
            for pos in range(1, len(segments)):
                segment = segments[pos]
                if segment[0] > current[1] + 1:
                    exons.append(current)
                    current = segment
                elif segment[0] == current[1] + 1:
                    current = (current[0], segment[1], None)
                else:
                    logger.warning("Overlapping segments found in %s. Discarding it", tid)
                    return []
            exons.append(current)
        tobject["features"]["exon"] = exons[:]
    else:
        raise KeyError(tobject["features"])

    segments = sorted(segments, key=itemgetter(0))
    tlength = 0
    start, end = segments[0][0], segments[-1][1]
    introns = []
    num_segments = len(segments)
    for pos, segment in enumerate(segments):
        if pos < num_segments - 1:
            later = segments[pos + 1]
            intron = later[0] - (segment[1] + 1)
            introns.append((pos, intron))
        tlength += segment[1] + 1 - segment[0]

    # Discard transcript under a certain size
    if tlength < min_length:
        if tobject["is_reference"] is True:
            logger.info("%s retained even if it is too short (%d) as it is a reference transcript.",
                        tid, tlength)
        else:
            logger.info("Discarding %s because its size (%d) is under the minimum of %d",
                        tid, tlength, min_length)
            return []

    # Discard transcripts with introns over the limit
    over = [intron for intron in introns if intron[1] > max_intron]
    if len(over) > 0:
        if tobject["is_reference"] is True:
            logger.info(
                "%s retained even if has %s introns the limit (%d, max: %d) as it is a reference transcript.",
                tid, len(over), max([_[1] for _ in over]), max_intron)
            return [(tobject, tid, start, end)]
        else:
            logger.info(
                "Splitting %s into %d transcripts because it has %d introns over the maximum of %d (longest: %d)",
                tid, len(over) + 1, len(over), max_intron, max([_[1] for _ in over]))
            splitted = []
            current = 0
            for num, ointron in enumerate(over):
                final_pos = ointron[0]
                segs = segments[current:final_pos+1][:]
                current = final_pos + 1
                start, stop = segs[0][0], segs[-1][1]
                tlength = sum([_[1] + 1 - _[0] for _ in segs])
                if tlength < min_length:
                    logger.info("Discarding fragment %s of %s because its length is beneath the minimum of %s (%s)",
                                num, tid, min_length, tlength)
                    continue
                else:
                    splitted.append(_create_split_tobject(tobject, start, stop, num))

            segs = segments[current:]
            start, stop = segs[0][0], segs[-1][1]
            tlength = sum([_[1] + 1 - _[0] for _ in segs])
            if tlength < min_length:
                logger.info("Discarding fragment %s of %s because its length is beneath the minimum of %s (%s)",
                            len(over), tid, min_length, tlength)
            else:
                splitted.append(_create_split_tobject(tobject, start, stop, len(over)))
            return splitted
    else:
        return [(tobject, tid, start, end)]


def load_into_storage(shelf_name, exon_lines, min_length, logger, strip_cds=True, max_intron=3*10**5):

    """Function to load the exon_lines dictionary into the temporary storage."""

    if os.path.exists(shelf_name) or any(_.startswith(os.path.basename(shelf_name))
                                         for _ in os.listdir(os.path.dirname(shelf_name))):
        logger.error("Shelf %s already exists (maybe from a previous aborted run?), dropping its contents", shelf_name)
        for _ in (_ for _ in os.listdir(os.path.dirname(shelf_name))
                  if _.startswith(os.path.basename(shelf_name))):
            if os.path.exists(_):
                os.remove(_)

    shelf = open(shelf_name, "wb")

    rows = []
    logger.warning("Max intron: %s", max_intron)
    for tid in exon_lines:
        if "features" not in exon_lines[tid]:
            raise KeyError("{0}: {1}\n{2}".format(tid, "features", exon_lines[tid]))
        if ("exon" not in exon_lines[tid]["features"] or
                len(exon_lines[tid]["features"]["exon"]) == 0):
            # Match-like things
            if "match" in exon_lines[tid]["features"]:
                if len(exon_lines[tid]["features"]["match"]) > 1:
                    logger.warning("Invalid features for %s, skipping.", tid)
                    continue
                exon_lines[tid]["features"]["exon"] = [exon_lines[tid]["features"]["match"][0]]
                logger.warning("Inferring that %s is a mono-exonic transcript-match: (%s, %d-%d)",
                               tid, exon_lines[tid]["chrom"],
                               exon_lines[tid]["features"]["exon"][0][0],
                               exon_lines[tid]["features"]["exon"][0][1])
                del exon_lines[tid]["features"]["match"]
            elif (strip_cds is False and "CDS" in exon_lines[tid]["features"] and
                len(exon_lines[tid]["features"]["CDS"]) > 0):
                pass
            else:
                logger.warning("No valid exon feature for %s, continuing", tid)
                continue
        elif "match" in exon_lines[tid]["features"] and "exon" in exon_lines[tid]["features"]:
            del exon_lines[tid]["features"]["match"]

        for values, tid, start, end in _evaluate_tid(tid, exon_lines[tid], logger,
                                                     max_intron=max_intron,
                                                     min_length=min_length):
            chrom = values["chrom"]
            assert chrom is not None
            strand = values["strand"]
            if strand is None:
                strand = "."

            logger.debug("Inserting %s into shelf %s", tid, shelf_name)
            values = zlib.compress(msgpack.dumps(values))
            write_start = shelf.tell()
            write_length = shelf.write(values)
            row = (chrom.encode(), start, end, strand.encode(), tid.encode(), write_start, write_length)
            rows.append(row)

    logger.warning("Finished packing rows for %s", shelf_name)
    return rows


def load_from_gff(shelf_name,
                  gff_handle,
                  label,
                  found_ids,
                  logger,
                  min_length=0,
                  max_intron=3*10**5,
                  is_reference=False,
                  exclude_redundant=False,
                  strip_cds=False,
                  strand_specific=False):
    """
    Method to load the exon lines from GFF3 files.
    :param shelf_name: the name of the shelf DB to use.
    :param gff_handle: The handle for the GTF to be parsed.
    :param label: label to be attached to all transcripts.
    :type label: str
    :param found_ids: set of IDs already found in other files.
    :type found_ids: set
    :param logger: a logger to be used to pass messages
    :type logger: logging.Logger
    :param min_length: minimum length for a cDNA to be considered as valid
    :type min_length: int
    :param max_intron: maximum intron length for a cDNA to be considered as valid
    :type max_intron: int
    :param strip_cds: boolean flag. If true, all CDS lines will be ignored.
    :type strip_cds: bool
    :param strand_specific: whether the assembly is strand-specific or not.
    :type strand_specific: bool
    :param is_reference: boolean. If set to True, the transcript will always be retained.
    :type is_reference: bool
    :param exclude_redundant: boolean. If set to True, fully redundant transcripts will be removed.
    :type exclude_redundant: bool
    :return:
    """

    exon_lines = dict()

    strip_cds = strip_cds and (not is_reference)
    if strand_specific is not True and is_reference is True:
        strand_specific = True

    transcript2genes = dict()
    new_ids = set()

    to_ignore = set()

    for row in gff_handle:
        if row.feature == "protein":
            continue
        elif row.is_transcript is True or row.feature == "match":
            if label != '':
                row.id = "{0}_{1}".format(label, row.id)
                row.source = label
            if row.id in found_ids:
                __raise_redundant(row.id, gff_handle.name, label)
            elif row.id in exon_lines:
                # This might sometimes happen in GMAP
                logger.warning(
                    "Multiple instance of %s found, skipping any subsequent entry",
                    row.id)
                to_ignore.add(row.id)
                continue
            #
            # if row.id not in exon_lines:
            exon_lines[row.id] = dict()
            exon_lines[row.id]["source"] = row.source
            if row.parent:
                transcript2genes[row.id] = row.parent[0]
            else:
                transcript2genes[row.id] = row.id
            assert row.id is not None
            if row.id in found_ids:
                __raise_redundant(row.id, gff_handle.name, label)

            exon_lines[row.id]["attributes"] = row.attributes.copy()
            exon_lines[row.id]["chrom"] = row.chrom
            exon_lines[row.id]["strand"] = row.strand
            exon_lines[row.id]["tid"] = row.transcript or row.id
            exon_lines[row.id]["parent"] = "{}.gene".format(row.id)
            exon_lines[row.id]["features"] = dict()
            # Here we have to add the match feature as an exon, in case it is the only one present
            if row.feature == "match":
                exon_lines[row.id]["features"][row.feature] = []
                exon_lines[row.id]["features"][row.feature].append((row.start, row.end, row.phase))

            exon_lines[row.id]["strand_specific"] = strand_specific
            exon_lines[row.id]["is_reference"] = is_reference
            exon_lines[row.id]["exclude_redundant"] = exclude_redundant
            continue
        elif row.is_exon is True:
            if not row.is_cds or (row.is_cds is True and strip_cds is False):
                if len(row.parent) == 0 and "cDNA_match" == row.feature:
                    if label == '':
                        __tid = row.id
                    else:
                        __tid = "{0}_{1}".format(label, row.id)
                    row.parent = __tid
                    transcript2genes[__tid] = "{}_match".format(__tid)
                    row.feature = "exon"
                elif row.feature == "match_part":
                    if label == '':
                        __tid = row.parent[0]
                    else:
                        __tid = "{0}_{1}".format(label, row.parent[0])
                    row.parent = __tid
                    transcript2genes[__tid] = "{}_match".format(__tid)
                    row.feature = "exon"

                elif label != '':
                    row.transcript = ["{0}_{1}".format(label, tid) for tid in row.transcript]

                parents = row.transcript[:]
                for tid in parents:

                    if tid in found_ids:
                        __raise_redundant(tid, gff_handle.name, label)
                    elif tid in to_ignore:
                        continue
                    if tid not in exon_lines and tid in transcript2genes:
                        exon_lines[tid] = dict()
                        exon_lines[tid]["attributes"] = row.attributes.copy()
                        if label:
                            exon_lines[tid]["source"] = label
                        else:
                            exon_lines[tid]["source"] = row.source
                        exon_lines[tid]["chrom"] = row.chrom
                        exon_lines[tid]["strand"] = row.strand
                        exon_lines[tid]["features"] = dict()
                        exon_lines[tid]["tid"] = tid
                        exon_lines[tid]["parent"] = transcript2genes[tid]
                        exon_lines[tid]["strand_specific"] = strand_specific
                        exon_lines[tid]["is_reference"] = is_reference
                        exon_lines[tid]["exclude_redundant"] = exclude_redundant
                    elif tid not in exon_lines and tid not in transcript2genes:
                        continue
                    else:
                        if "exon_number" in row.attributes:
                            del row.attributes["exon_number"]
                        if (exon_lines[tid]["chrom"] != row.chrom or
                                exon_lines[tid]["strand"] != row.strand):
                            __raise_invalid(tid, gff_handle.name, label)
                        exon_lines[tid]["attributes"].update(row.attributes)

                    if row.feature not in exon_lines[tid]["features"]:
                        exon_lines[tid]["features"][row.feature] = []
                    exon_lines[tid]["features"][row.feature].append((row.start, row.end, row.phase))
                    new_ids.add(tid)

            else:
                continue
    gff_handle.close()

    logger.info("Starting to load %s", shelf_name)
    rows = load_into_storage(shelf_name, exon_lines,
                      logger=logger, min_length=min_length, strip_cds=strip_cds, max_intron=max_intron)

    logger.info("Finished parsing %s", gff_handle.name)
    return new_ids, rows


def load_from_gtf(shelf_name,
                  gff_handle,
                  label,
                  found_ids,
                  logger,
                  min_length=0,
                  max_intron=3*10**5,
                  is_reference=False,
                  exclude_redundant=False,
                  strip_cds=False,
                  strand_specific=False):
    """
    Method to load the exon lines from GTF files.
    :param shelf_name: the name of the shelf DB to use.
    :param gff_handle: The handle for the GTF to be parsed.
    :param label: label to be attached to all transcripts.
    :type label: str
    :param found_ids: set of IDs already found in other files.
    :type found_ids: set
    :param logger: a logger to be used to pass messages
    :type logger: logging.Logger
    :param min_length: minimum length for a cDNA to be considered as valid
    :type min_length: int
    :param max_intron: maximum intron length for a cDNA to be considered as valid
    :type max_intron: int
    :param strip_cds: boolean flag. If true, all CDS lines will be ignored.
    :type strip_cds: bool
    :param strand_specific: whether the assembly is strand-specific or not.
    :type strand_specific: bool
    :param is_reference: boolean. If set to True, the transcript will always be retained.
    :type is_reference: bool
    :param exclude_redundant: boolean. If set to True, the transcript will be marked for potential redundancy removal.
    :type exclude_redundant: bool
    :return:
    """

    exon_lines = dict()

    strip_cds = strip_cds and (not is_reference)
    strand_specific = strand_specific or is_reference

    # Reduce memory footprint
    [intern(_) for _ in ["chrom", "features", "strand", "attributes", "tid", "parent", "attributes"]]

    new_ids = set()
    to_ignore = set()
    for row in gff_handle:
        if row.is_transcript is True:
            if label != '':
                row.transcript = "{0}_{1}".format(label, row.transcript)
            if row.transcript in found_ids:
                __raise_redundant(row.transcript, gff_handle.name, label)
            if row.transcript in exon_lines:
                logger.warning(
                    "Multiple instance of %s found, skipping any subsequent entry", row.id)
                to_ignore.add(row.id)
                continue
                # __raise_invalid(row.transcript, gff_handle.name, label)
            if row.transcript not in exon_lines:
                exon_lines[row.transcript] = dict()
            if label:
                exon_lines[row.transcript]["source"] = label
            else:
                exon_lines[row.transcript]["source"] = row.source

            exon_lines[row.transcript]["features"] = dict()
            exon_lines[row.transcript]["chrom"] = row.chrom
            exon_lines[row.transcript]["strand"] = row.strand
            exon_lines[row.transcript]["attributes"] = row.attributes.copy()
            exon_lines[row.transcript]["tid"] = row.id
            exon_lines[row.transcript]["parent"] = "{}.gene".format(row.id)
            exon_lines[row.transcript]["strand_specific"] = strand_specific
            exon_lines[row.transcript]["is_reference"] = is_reference
            exon_lines[row.transcript]["exclude_redundant"] = exclude_redundant
            if "exon_number" in exon_lines[row.transcript]["attributes"]:
                del exon_lines[row.transcript]["attributes"]["exon_number"]
            continue

        if row.is_exon is False or (row.is_cds is True and strip_cds is True):
            continue
        if label != '':
            row.transcript = "{0}_{1}".format(label, row.transcript)
        if row.transcript in found_ids:
            __raise_redundant(row.transcript, gff_handle.name, label)
        assert row.transcript is not None
        if row.transcript not in exon_lines:
            exon_lines[row.transcript] = dict()
            if label:
                exon_lines[row.transcript]["source"] = label
            else:
                exon_lines[row.transcript]["source"] = row.source
            exon_lines[row.transcript]["features"] = dict()
            exon_lines[row.transcript]["chrom"] = row.chrom
            exon_lines[row.transcript]["strand"] = row.strand
            exon_lines[row.transcript]["exon"] = []
            exon_lines[row.transcript]["attributes"] = row.attributes.copy()
            exon_lines[row.transcript]["tid"] = row.transcript
            exon_lines[row.transcript]["parent"] = "{}.gene".format(row.transcript)
            exon_lines[row.transcript]["strand_specific"] = strand_specific
            exon_lines[row.transcript]["is_reference"] = is_reference
            exon_lines[row.transcript]["exclude_redundant"] = exclude_redundant
        else:
            if row.transcript in to_ignore:
                continue
            if "exon_number" in row.attributes:
                del row.attributes["exon_number"]
            if ("chrom" not in exon_lines[row.transcript] or
                    exon_lines[row.transcript]["chrom"] != row.chrom or
                    exon_lines[row.transcript]["strand"] != row.strand):
                __raise_invalid(row.transcript, gff_handle.name, label)
            exon_lines[row.transcript]["attributes"].update(row.attributes)
        if row.feature not in exon_lines[row.transcript]["features"]:
            exon_lines[row.transcript]["features"][row.feature] = []
        exon_lines[row.transcript]["features"][row.feature].append((row.start, row.end, row.phase))
        new_ids.add(row.transcript)
    gff_handle.close()
    logger.info("Starting to load %s", shelf_name)
    rows = load_into_storage(shelf_name,
                      exon_lines,
                      logger=logger, min_length=min_length, strip_cds=strip_cds, max_intron=max_intron)

    logger.info("Finished parsing %s", gff_handle.name)
    return new_ids, rows


def load_from_bed12(shelf_name,
                    gff_handle,
                    label,
                    found_ids,
                    logger,
                    min_length=0,
                    max_intron=3*10**5,
                    is_reference=False,
                    exclude_redundant=False,
                    strip_cds=False,
                    strand_specific=False):
    """
    Method to load the exon lines from GTF files.
    :param shelf_name: the name of the shelf DB to use.
    :param gff_handle: The handle for the GTF to be parsed.
    :param label: label to be attached to all transcripts.
    :type label: str
    :param found_ids: set of IDs already found in other files.
    :type found_ids: set
    :param logger: a logger to be used to pass messages
    :type logger: logging.Logger
    :param min_length: minimum length for a cDNA to be considered as valid
    :type min_length: int
    :param max_intron: maximum intron length for a cDNA to be considered as valid
    :type max_intron: int
    :param strip_cds: boolean flag. If true, all CDS lines will be ignored.
    :type strip_cds: bool
    :param strand_specific: whether the assembly is strand-specific or not.
    :type strand_specific: bool
    :param is_reference: boolean. If set to True, the transcript will always be retained.
    :type is_reference: bool
    :param exclude_redundant: boolean. If set to True, the transcript will be marked for potential redundancy removal.
    :type exclude_redundant: bool
    :return:
    """

    exon_lines = dict()

    strip_cds = strip_cds and (not is_reference)
    strand_specific = strand_specific or is_reference

    # Reduce memory footprint
    [intern(_) for _ in ["chrom", "features", "strand", "attributes", "tid", "parent", "attributes"]]

    new_ids = set()
    to_ignore = set()
    for row in gff_handle:
        # Each row is a transcript
        transcript = Transcript(row)
        if label != '':
            transcript.id = "{0}_{1}".format(label, transcript.id)
            if transcript.id in found_ids:
                __raise_redundant(transcript.id, gff_handle.name, label)
            if transcript.id in exon_lines:
                logger.warning(
                    "Multiple instance of %s found, skipping any subsequent entry", row.id)
                to_ignore.add(row.id)
                continue
            else:
                exon_lines[transcript.id] = dict()
            if label:
                exon_lines[transcript.id]["source"] = label
            else:
                exon_lines[transcript.id]["source"] = gff_handle.name  # BED12 files have no source
            exon_lines[transcript.id]["features"] = dict()
            exon_lines[transcript.id]["chrom"] = transcript.chrom
            exon_lines[transcript.id]["strand"] = transcript.strand
            # Should deal with GFFRead style input and BAM
            exon_lines[transcript.id]["attributes"] = transcript.attributes
            exon_lines[transcript.id]["tid"] = transcript.id
            exon_lines[transcript.id]["parent"] = "{}.gene".format(transcript.id)
            exon_lines[transcript.id]["strand_specific"] = strand_specific
            exon_lines[transcript.id]["is_reference"] = is_reference
            exon_lines[transcript.id]["exclude_redundant"] = exclude_redundant
            exon_lines[transcript.id]["features"]["exon"] = [
                (exon[0], exon[1]) for exon in transcript.exons
            ]
            if transcript.is_coding and not strip_cds:
                exon_lines[transcript.id]["features"]['CDS'] = [
                    (exon[0], exon[1]) for exon in transcript.combined_cds
                ]
                exon_lines[transcript.id]["features"]["UTR"] = [
                    (exon[0], exon[1]) for exon in transcript.five_utr + transcript.three_utr
                ]
        new_ids.add(transcript.id)
    gff_handle.close()
    rows = load_into_storage(shelf_name, exon_lines,
                      logger=logger, min_length=min_length, strip_cds=strip_cds, max_intron=max_intron)

    logger.info("Finished parsing %s", gff_handle.name)
    return new_ids, rows


def load_from_bam(shelf_name: str,
                  gff_handle: BamParser,
                  label: str,
                  found_ids: set,
                  logger: logging.Logger,
                  min_length=0,
                  max_intron=3*10**5,
                  is_reference=False,
                  exclude_redundant=False,
                  strip_cds=False,
                  strand_specific=False):
    """
    Method to load the exon lines from BAM files.
    :param shelf_name: the name of the shelf DB to use.
    :param gff_handle: The handle for the BAM to be parsed. This handle is BamParser with a file attached to read from.
    :param label: label to be attached to all transcripts.
    :type label: str
    :param found_ids: set of IDs already found in other files.
    :type found_ids: set
    :param logger: a logger to report any messages
    :type logger: logging.Logger
    :param min_length: minimum length for a cDNA to be considered as valid
    :type min_length: int
    :param max_intron: maximum intron length for a cDNA to be considered as valid
    :type max_intron: int
    :param strip_cds: boolean flag. If true, all CDS lines will be ignored.
    :type strip_cds: bool
    :param strand_specific: whether the input data is strand-specific or not.
    :type strand_specific: bool
    :param is_reference: boolean. If set to True, the transcript will always be retained.
    :type is_reference: bool
    :param exclude_redundant: boolean. If set to True, the transcript will be marked for potential redundancy removal.
    :type exclude_redundant: bool
    :return:
    """
    return load_from_bed12(shelf_name, gff_handle, label, found_ids, logger,
                           min_length=min_length, max_intron=max_intron,
                           is_reference=is_reference,exclude_redundant=exclude_redundant,
                           strip_cds=strip_cds, strand_specific=strand_specific)


loaders = {"gtf": load_from_gtf, "gff": load_from_gff, "gff3": load_from_gff,
           "bed12": load_from_bed12, "bed": load_from_bed12, "bam": load_from_bam}

# Chrom, start, end, strand, Tid, write start, write length
# 100 chars, unsigned Long, unsigned Long, one char, 100 chars, unsigned Long, unsigned Long
_row_struct_str = ">1000sLLc1000sLLH"
row_struct = struct.Struct(_row_struct_str)
row_struct_size = struct.calcsize(_row_struct_str)


class AnnotationParser(multiprocessing.Process):

    def __init__(self,
                 submission_queue: multiprocessing.JoinableQueue,
                 return_queue: multiprocessing.JoinableQueue,
                 logging_queue: multiprocessing.JoinableQueue,
                 identifier: int,
                 min_length=0,
                 max_intron=3*10**5,
                 log_level="WARNING",
                 seed=None,
                 strip_cds=False):

        super().__init__()
        if seed is not None:
            # numpy.random.seed(seed % (2 ** 32 - 1))
            random.seed(seed % (2 ** 32 - 1))
        else:
            # numpy.random.seed(None)
            random.seed(None)

        self.submission_queue = submission_queue
        self.return_queue = return_queue
        self.min_length = min_length
        self.max_intron = max_intron
        self.__strip_cds = strip_cds
        self.logging_queue = logging_queue
        self.log_level = log_level
        self.__identifier = identifier
        self.name = "AnnotationParser-{0}".format(self.identifier)
        self.logger = None
        self.handler = None
        self.logger = logging.getLogger(self.name)
        create_queue_logger(self, prefix="prepare")
        # self.logger.warning("Started process %s", self.name)

    def __getstate__(self):

        state = self.__dict__.copy()
        for key in ("logger", "handler", "_log_handler"):
            if key in state:
                del state[key]

        return state

    def __setstate__(self, state):

        self.__dict__.update(state)
        create_queue_logger(self)

    def run(self):

        found_ids = set()
        self.logger.debug("Starting to listen to the queue")
        counter = 0
        while True:
            results = self.submission_queue.get()
            try:
                label, handle, strand_specific, is_reference,\
                exclude_redundant, file_strip_cds, shelf_name, shelf_index = results
            except ValueError as exc:
                raise ValueError("{}.\tValues: {}".format(exc, ", ".join([str(_) for _ in results])))
            if handle == "EXIT":
                self.submission_queue.put(results)
                break
            counter += 1
            self.logger.debug("Received %s (label: %s; SS: %s, shelf_name: %s)",
                              handle,
                              label,
                              strand_specific,
                              shelf_name)
            try:
                gff_handle = parser_factory(handle)
                loader = loaders.get(gff_handle.__annot_type__, None)
                if loader is None:
                    raise ValueError("Invalid file type: {}".format(gff_handle.name))
                if file_strip_cds is True:
                    file_strip_cds = True
                else:
                    file_strip_cds = self.__strip_cds

                new_ids, new_rows = loader(shelf_name, gff_handle, label, found_ids, self.logger,
                                           min_length=self.min_length, max_intron=self.max_intron,
                                           strip_cds=file_strip_cds and not is_reference,
                                           is_reference=is_reference, exclude_redundant=exclude_redundant,
                                           strand_specific=strand_specific)

                if len(new_ids) == 0:
                    raise exceptions.InvalidAssembly(
                        "No valid transcripts found in {0}{1}!".format(
                            handle, " (label: {0})".format(label) if label != "" else ""
                        ))
                # Now convert the rows into structs.
                self.logger.debug("Packing %d rows of %s", len(new_rows), label)
                [self.return_queue.put_nowait((*row, shelf_index)) for row in new_rows]
                self.logger.debug("Packed %d rows of %s", len(new_rows), label)

            except (exceptions.InvalidAssembly, exceptions.InvalidParsingFormat) as exc:
                self.logger.exception("Invalid file: %s. Skipping it", handle)
                self.logger.exception(exc)
                load_into_storage(shelf_name, [], self.min_length, self.logger, strip_cds=True,
                                                      max_intron=3 * 10 ** 5)
                [self.return_queue.put_nowait((*row, shelf_index)) for row in []]
                continue
            except Exception as exc:
                self.logger.exception(exc)
                raise

        self.return_queue.put_nowait("FINISHED")

    @property
    def identifier(self):
        """
        A numeric value that identifies the process uniquely.
        :return:
        """
        return self.__identifier



from dataclasses import field
from typing import List
from marshmallow import validate
from marshmallow_dataclass import dataclass, Optional


@dataclass
class PrepareFilesConfiguration:
    output_dir: str = field(default=".", metadata={"metadata": {"description": "Output folder."},
                                                   "validate": validate.Length(min=1)})
    out: str = field(default="mikado_prepared.gtf", metadata={
        "metadata": {"description": "Output GTF file."},
        "validate": validate.Length(min=1)}
    )
    out_fasta: str = field(default="mikado_prepared.fasta", metadata={
        "validate": validate.Length(min=1)
    })
    log: str = field(default="prepare.log", metadata={
        "metadata": {"description": "Log file."}
    })
    gff: List[str] = field(default_factory=lambda: [], metadata={
        "metadata": {"description": "List of the input files."},
        "required": True
    })
    labels: List[str] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "List of labels associated with each input GFF. This list must *either* be empty *or* \
be the same length as the 'gff' content above."},
    })
    strand_specific_assemblies: List[str] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "List of input assemblies to be considered as strand specific. Any 'reference' \
input is automatically marked as strand-specific."},
        "required": True
    })
    reference: List[bool] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "List of input assemblies to be considered as of 'reference' quality. Transcripts from this \
list will be excluded only if they have glaring mistakes eg. an incorrect CDS."},
    })
    exclude_redundant: List[bool] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "If the 'exclude_redundant' switch is set to true, this list specifies which assemblies \
can have the redundancy check performed on."},
    })
    strip_cds: List[bool] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "List of input assemblies from which the CDS will be stripped. Useful to e.g. remove \
the CDS from GMAP alignments."},
    })
    source_score: dict = field(default_factory=dict, metadata={
        "metadata": {
            "description": "Dictionary linking assemblies with an optional score, for prioritising them during \
'pick' (or de-prioritising if the score is negative)."},
    })


@dataclass
class PrepareConfiguration:
    exclude_redundant: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean. If set to True, fully redundant transcripts across the input files will be \
removed. Higher scoring transcripts *on the basis of the score associated to a given input file* will be \
preferentially retained."}
    })
    minimum_cdna_length: int = field(default=200, metadata={
        "metadata": {"description": "Minimum length of a transcript to be retained. Default: 200 bps"},
                     "validate": validate.Range(min=1),
                     "required": True
    })
    max_intron_length: int = field(default=1000000, metadata={
        "metadata": {
            "description": "Maximum length of an intron. Transcripts with introns bigger than this value will be \
split in various sub-transcripts. Default: 1,000,000 bps."},
            "validate": validate.Range(min=20),
            "required": True
    })
    strip_cds: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set, the CDS will be stripped from any non-reference input assembly."},
    })
    strip_faulty_cds: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set to false, any transcript - *including reference transcripts* found \
to have an incorrect CDS will be discarded. If set to to true, these transcripts will be retained \
but their CDS will be stripped."},
    })
    single: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set to true, Mikado will run in single-threaded mode, \
useful for debugging."},
    })
    lenient: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set to true, Mikado will retain transcripts with no canonical junction \
or with canonical junctions on both strands. If set to false (default), such transcripts will instead be discarded."},
    })
    strand_specific: bool = field(default=False, metadata={
        "metadata": {
            "description": "Boolean flag. If set to true, all assemblies will be considered as strand-specific. \
By default, Mikado will consider the strand-specificity of each assembly in isolation, \
see 'files/strand_specific_assemblies'."},
            "required": True
    })
    canonical: List[List[str]] = field(default_factory=lambda: [["GT", "AG"], ["GC", "AG"], ["AT", "AC"]], metadata={
        "metadata": {"description": "Accepted canonical splicing junctions for the organism in examination."},
    })
    files: Optional[PrepareFilesConfiguration] = field(default_factory=PrepareFilesConfiguration, metadata={
        "metadata": {"description": "Options related to the input and output files."},
                     "required": True
    })



#!/usr/bin/env python3
# coding=utf-8

"""
This module defines the Picker class, which is the main workhorse for Mikado pick.
"""

import sys
import re
import csv
import os
import shutil
import tempfile
import random
import logging
import zlib
from logging import handlers as logging_handlers
import functools
import multiprocessing

from Mikado import version
from sqlalchemy.engine import create_engine  # SQLAlchemy/DB imports
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.pool import QueuePool as SqlPool
import sqlalchemy
import sqlalchemy.exc
from ..utilities import path_join, create_shm_handle
from ..utilities.log_utils import formatter
from ..parsers.GTF import GTF, GtfLine
from ..parsers.GFF import GFF3
from ..parsers.bed12 import Bed12Parser
from ..parsers import Parser
from ..serializers.junction import Chrom
from ..serializers.external import ExternalSource
from ..transcripts import Transcript
from ..loci.superlocus import Superlocus
from ..configuration.configurator import load_and_validate_config
from ..utilities import dbutils
from ..exceptions import UnsortedInput, InvalidConfiguration, InvalidTranscript, InvalidCDS
from .loci_processer import analyse_locus, LociProcesser, merge_loci
from ._locus_single_printer import print_locus
import multiprocessing.managers
import warnings
import pyfaidx
import msgpack
from numpy import percentile
import dataclasses
from ..configuration.configuration import MikadoConfiguration
from ..configuration.daijin_configuration import DaijinConfiguration
logging.captureWarnings(True)
warnings.simplefilter("always")
import rapidjson as json


# pylint: disable=too-many-instance-attributes
class Picker:

    """
    This class is used to launch the main Mikado pipeline. Its purpose is to parse
    an input sorted annotation file, locate the loci, and perform the selection analysis
    using the parameters provided in the input configuration file.
    """

    # @profile
    def __init__(self, configuration, commandline="", regions=None):

        """Constructor. It takes a single argument as input - the JSON/YAML configuration,
        prepared by the json_utils functions.

        :param configuration: Either a configuration dictionary or the configuration file.
        :type configuration: (MikadoConfiguration|DaijinConfiguration)

        :param commandline: optional, the commandline used to start the program
        :type commandline: str
        """

        # Mock variables
        self.formatter = self.main_logger = self.log_writer = self.log_handler = self.logger = None
        self.log_level = "WARN"

        # # Things that have to be deleted upon serialisation
        # Now we start the real work
        self.commandline = commandline
        self.configuration = configuration

        self.__load_configuration()
        self.shm_db = None
        self.setup_shm_db()
        self.__regions = regions
        self.procs = self.configuration.threads

        # Check the input file
        with self.define_input() as _:
            pass

        self.__create_output_handles()
        # pylint: disable=no-member
        multiprocessing.set_start_method(self.configuration.multiprocessing_method,
                                         force=True)

        # self.setup_logger()
        self.logger.info("Starting to analyse input file %s", self.input_file)
        self.logger.info("Random seed: %s", self.configuration.seed)
        random.seed((self.configuration.seed) % (2 ** 32 - 1))
        self.logger.debug("Multiprocessing method: %s", self.configuration.multiprocessing_method)
        # pylint: enable=no-member
        self.manager = self.context.Manager()
        self.db_connection = functools.partial(
            dbutils.create_connector,
            self.configuration,
            self.logger)

        # Update the database if necessary
        engine = create_engine("{0}://".format(self.configuration.db_settings.dbtype), creator=self.db_connection)
        dbutils.DBBASE.metadata.create_all(engine)
        engine.dispose()

        if self.configuration.pick.run_options.single_thread is True:
            # Reset threads to 1
            if self.configuration.threads > 1:
                self.main_logger.warning("Reset number of threads to 1 as requested")
                self.procs = 1

        if self.locus_out is None:
            raise InvalidConfiguration(
                "No output prefix specified for the final loci. Key: \"loci_out\"")

        # self.printer_process = Process(target=self.printer)
        self.queue_pool = None

    def define_input(self, multithreading=False):
        """Function to check that the input file exists and is valid. It returns the parser."""

        if ".gtf" in self.input_file:
            parser = GTF
            fformat = "gtf"
        elif ".gff" in self.input_file:
            parser = GFF3
            fformat = "gff3"
        elif ".bed" in self.input_file:
            parser = Bed12Parser
            fformat = "bed12"
        else:
            raise InvalidConfiguration("Invalid input file: {0}".format(self.input_file))

        verified = False
        with parser(self.input_file) as testing:
            for row in testing:
                if row.header is False:
                    verified = True
                    break
        if verified is False:
            raise InvalidConfiguration("Invalid input file: {0}".format(self.input_file))

        if multithreading:
            parser = Parser(self.input_file)
            parser.format = fformat
            return parser
        else:
            return parser(self.input_file)

    def __load_configuration(self):

        """Private method to load the configuration"""

        if isinstance(self.configuration, str):
            assert os.path.exists(self.configuration)
            self.configuration = load_and_validate_config(self.configuration, logger=self.logger)
            assert isinstance(self.configuration, (MikadoConfiguration, DaijinConfiguration))
            # pylint: disable=no-member
        elif not isinstance(self.configuration, (MikadoConfiguration,DaijinConfiguration)):
            raise TypeError(type(self.configuration))

        multiprocessing.set_start_method(self.configuration.multiprocessing_method,
                                         force=True)
        self.input_file = self.configuration.pick.files.input
        self.setup_logger()

        if self.configuration.pick.alternative_splicing.pad is True:
            # Check that, when asks for padding, the reference genome is present
            self.logger.debug("Checking for the presence of the reference genome")
            try:
                _ = pyfaidx.Fasta(self.configuration.reference.genome)
            except (pyfaidx.FastaIndexingError, FileNotFoundError, pyfaidx.FastaNotFoundError):
                self.logger.error("Transcript padding cannot be executed without a valid genome file.\
                 Please, either disable the padding or provide a valid genome sequence.")
                sys.exit(1)
            self.logger.debug("Valid reference genome found")
        else:
            pass

        self.context = multiprocessing.get_context()
        self.logger.debug("Configuration loaded successfully")

    def __create_output_handles(self):

        """Create all the output-related variables."""

        if self.configuration.pick.files.subloci_out:
            self.sub_out = path_join(
                self.configuration.pick.files.output_dir,
                self.configuration.pick.files.subloci_out
            )
        else:
            self.sub_out = ""
        if self.configuration.pick.files.monoloci_out:
            self.monolocus_out = path_join(
                self.configuration.pick.files.output_dir,
                self.configuration.pick.files.monoloci_out
            )
        else:
            self.monolocus_out = ""
        self.locus_out = path_join(
            self.configuration.pick.files.output_dir,
            self.configuration.pick.files.loci_out)

        assert self.locus_out != ''
        assert self.locus_out != self.sub_out and self.locus_out != self.monolocus_out
        assert (not self.sub_out and not self.monolocus_out) or (self.sub_out != self.monolocus_out)

    def setup_shm_db(self):
        """
        This method will copy the SQLite input DB into memory.
        """

        if self.configuration.db_settings.dbtype == "sqlite" and os.path.exists(self.configuration.db_settings.db):
            if self.configuration.pick.run_options.shm is True:
                handle = create_shm_handle()
                if handle is None:
                    self.main_logger.info("Mikado was asked to copy the database into /dev/shm, but it is either "
                                          "not available or not writable for this user.")
                    handle = tempfile.NamedTemporaryFile(suffix=".db")
            else:
                handle = tempfile.NamedTemporaryFile(suffix=".db")
            self.shm_db = handle
            shutil.copy2(self.configuration.db_settings.db, self.shm_db.name)
            self.configuration.db_settings.db = self.shm_db.name
            self.main_logger.info(f"DB copied into {self.shm_db.name}")

    def setup_logger(self):

        """This function sets up the logger for the class.
        It creates the instance attribute "log_writer", which is itself a
        logging.handlers.QueueListener instance listening on the logging_queue
        instance attribute (which is a normal mp.Manager.Queue instance)."""

        self.logging_queue = multiprocessing.Queue(-1)
        self.printer_queue = multiprocessing.Queue(-1)
        self.formatter = formatter
        self.main_logger = logging.getLogger("main_logger")
        if not os.path.exists(self.configuration.pick.files.output_dir):
            try:
                os.makedirs(self.configuration.pick.files.output_dir)
            except (OSError, PermissionError) as exc:
                self.logger.error("Failed to create the output directory!")
                self.logger.exception(exc)
                raise
        elif not os.path.isdir(self.configuration.pick.files.output_dir):
            self.logger.error(
                "The specified output directory %s exists and is not a file; aborting",
                self.configuration.pick.files.output_dir)
            raise OSError("The specified output directory %s exists and is not a file; aborting" %
                          self.configuration.pick.files.output_dir)

        self.logger = logging.getLogger("listener")
        self.logger.propagate = False
        if (self.configuration.pick.files.log is None or
                self.configuration.pick.files.log in ("stream", "")):
            self.log_handler = logging.StreamHandler()
        else:
            if os.path.basename(self.configuration.pick.files.log) == self.configuration.pick.files.log:
                fname = path_join(self.configuration.pick.files.output_dir,
                                  self.configuration.pick.files.log)
            else:
                fname = self.configuration.pick.files.log

            self.log_handler = logging.FileHandler(filename=fname, mode='w')
            assert os.path.exists(fname)

        # For the main logger I want to keep it at the "INFO" level
        self.log_level = self.configuration.log_settings.log_level
        self.log_handler.setFormatter(self.formatter)
        self.logger.setLevel(self.log_level)
        self.logger.addHandler(self.log_handler)

        if self.log_level == "DEBUG" and self.configuration.threads > 1:
            self.main_logger.setLevel(logging.DEBUG)
            self.main_logger.warning(
                    "Due to a Python design bug, we have to force Mikado to go in single-threaded mode when debugging.")
            self.procs = self.configuration.threads = 1
        else:
            self.main_logger.setLevel(logging.INFO)
        self.main_logger.addHandler(self.log_handler)

        self.main_logger.info(f"Mikado version: {version.__version__}")
        if self.commandline != '':
            self.main_logger.info("Command line: {0}".format(self.commandline))
        else:
            self.main_logger.info(
                "Analysis launched directly, without using the launch script.")

        self.main_logger.info("Begun analysis of {0}".format(self.input_file))
        # Create the shared DB if necessary
        self.log_writer = logging_handlers.QueueListener(self.logging_queue, self.logger)
        self.log_writer.start()

        self.logger_queue_handler = logging_handlers.QueueHandler(self.logging_queue)
        self.queue_logger = logging.getLogger("parser")
        self.queue_logger.addHandler(self.logger_queue_handler)

        self.queue_logger.setLevel(logging.getLevelName(self.configuration.log_settings.log_level))
        self.logger.warning("Current level for queue: %s", logging.getLevelName(self.queue_logger.level))

        self.queue_logger.propagate = False

        # Configure SQL logging
        sqllogger = logging.getLogger("sqlalchemy.engine")
        sqllogger.setLevel(self.configuration.log_settings.sql_level)
        sqllogger.addHandler(self.logger_queue_handler)

        return

    def __print_gff_headers(self, locus_out, score_keys, metrics):
        """Private method to print the GFF headers of the output files.
        Moreover, it will determine whether to start output files for
        subloci and/or monoloci.
        Returns: [subloci_metrics_handle, subloci_scores_handle, subloci_handle],
         monosubloci_outfile
        """

        print('##gff-version 3', file=locus_out)
        engine = create_engine("{0}://".format(self.configuration.db_settings.dbtype),
                               creator=self.db_connection)
        session = sqlalchemy.orm.sessionmaker(bind=engine)()
        try:
            for chrom in session.query(Chrom).order_by(Chrom.name.asc()):
                print("##sequence-region {0} 1 {1}".format(chrom.name, chrom.length),
                      file=locus_out)
                locus_out.flush()
        except sqlalchemy.exc.OperationalError as _:
            self.logger.error("Empty database! Creating a mock one")
            self.configuration.db_settings.dbtype = "sqlite"
            self.__db = tempfile.NamedTemporaryFile(suffix=".db", mode="wb")
            self.configuration.db_settings.db = self.__db.name
            engine = create_engine("{0}://".format(self.configuration.db_settings.dbtype),
                                   creator=self.db_connection)
            session = sqlalchemy.orm.sessionmaker(bind=engine)()
            dbutils.DBBASE.metadata.create_all(engine)

        if self.sub_out != '':
            assert isinstance(self.sub_out, str)
            sub_metrics_file = open(re.sub("$", ".metrics.tsv",
                                    re.sub(".gff.?$", "", self.sub_out)), "w")
            sub_scores_file = open(re.sub("$", ".scores.tsv",
                                   re.sub(".gff.?$", "", self.sub_out)), "w")
            sub_metrics = csv.DictWriter(
                sub_metrics_file,
                metrics,
                extrasaction="ignore",
                delimiter="\t")
            sub_metrics.writeheader()
            sub_scores = csv.DictWriter(
                sub_scores_file, score_keys, delimiter="\t", extrasaction="ignore",)
            sub_scores.writeheader()
            # Not a very clean wya to do things ... attaching the handles as properties
            sub_scores.handle = sub_scores_file
            sub_scores.flush = sub_scores.handle.flush
            sub_scores.flush()
            sub_scores.close = sub_scores.handle.close
            sub_scores.name = sub_scores.handle.name
            sub_metrics.handle = sub_metrics_file
            sub_metrics.flush = sub_metrics.handle.flush
            sub_metrics.flush()
            sub_metrics.close = sub_metrics.handle.close
            sub_metrics.name = sub_metrics.handle.name
            sub_out = open(self.sub_out, 'w')
            print('##gff-version 3', file=sub_out)
            for chrom in session.query(Chrom).order_by(Chrom.name.desc()):
                print("##sequence-region {0} 1 {1}".format(chrom.name, chrom.length),
                      file=sub_out)
            sub_out.flush()
            sub_files = [sub_metrics,
                         sub_scores,
                         sub_out]
        else:
            sub_files = [None, None, None]

        if self.monolocus_out != '':
            mono_metrics_file = open(re.sub("$", ".metrics.tsv",
                                     re.sub(".gff.?$", "", self.monolocus_out)), "w")
            mono_scores_file = open(re.sub("$", ".scores.tsv",
                                    re.sub(".gff.?$", "", self.monolocus_out)), "w")          
            mono_metrics = csv.DictWriter(
                mono_metrics_file,
                metrics,
                extrasaction="ignore",
                delimiter="\t")
            mono_metrics.writeheader()
            mono_scores = csv.DictWriter(
                mono_scores_file, score_keys, delimiter="\t",
                extrasaction="ignore")
            mono_scores.writeheader()
            # Not a very clean wya to do things ... attaching the handles as properties
            mono_scores.handle = mono_scores_file
            mono_scores.flush = mono_scores.handle.flush
            mono_scores.flush()
            mono_scores.close = mono_scores.handle.close
            mono_scores.name = mono_scores.handle.name
            mono_metrics.handle = mono_metrics_file
            mono_metrics.flush = mono_metrics.handle.flush
            mono_metrics.flush()
            mono_metrics.close = mono_metrics.handle.close
            mono_metrics.name = mono_metrics.handle.name
            mono_out = open(self.monolocus_out, 'w')
            print('##gff-version 3', file=mono_out)
            for chrom in session.query(Chrom).order_by(Chrom.name.desc()):
                print("##sequence-region {0} 1 {1}".format(chrom.name, chrom.length),
                      file=mono_out)

            mono_out.flush()
            mono_files = [mono_metrics,
                          mono_scores,
                          mono_out]
        else:
            mono_files = [None] * 3

        session.close()
        engine.dispose()
        return sub_files, mono_files

    def __get_output_files(self):

        """
        Private method used by printer to prepare all the output files.
        :return: ((locus_metrics, locus_scores, locus_out),
                (sub_metrics, sub_scores, sub_out),
                mono_out)
                all of these are file handles
        """

        engine = create_engine("{0}://".format(self.configuration.db_settings.dbtype),
                               creator=self.db_connection)
        session = sqlalchemy.orm.sessionmaker(bind=engine)()

        score_keys = ["source_score"]
        __scores = set(self.configuration.scoring.scoring.keys())

        available_external_metrics = ["external.{}".format(_.source)
                                      for _ in session.query(ExternalSource.source).all()]

        requested_external = set()
        requested_external.update({param for param in self.configuration.scoring.requirements.parameters.keys()
                                   if param.startswith("external")})
        requested_external.update({param for param in self.configuration.scoring.not_fragmentary.parameters.keys()
                                   if param.startswith("external")})
        requested_external.update({param for param in self.configuration.scoring.cds_requirements.parameters.keys()
                                   if param.startswith("external")})
        requested_external.update({param for param in self.configuration.scoring.cds_requirements.parameters.keys()
                                   if param.startswith("external")})
        requested_external.update({param for param in self.configuration.scoring.scoring.keys()
                                   if param.startswith("external")})
        
        attribute_metrics = set()
        attribute_metrics.update({param for param in self.configuration.scoring.requirements.parameters.keys()
                                   if param.startswith("attributes")})
        attribute_metrics.update({param for param in self.configuration.scoring.not_fragmentary.parameters.keys()
                                   if param.startswith("attributes")})
        attribute_metrics.update({param for param in self.configuration.scoring.cds_requirements.parameters.keys()
                                   if param.startswith("attributes")})
        attribute_metrics.update({param for param in self.configuration.scoring.cds_requirements.parameters.keys()
                                   if param.startswith("attributes")})
        attribute_metrics.update({param for param in self.configuration.scoring.scoring.keys()
                                   if param.startswith("attributes")})

        # Check that the external scores are all present. If they are not, raise a warning.
        if requested_external - set(available_external_metrics):
            self.logger.error(
                ("The following external metrics, found in the scoring file, are not present in the database. " +
                 "Please check their existence:\n" + "\n".join(
                            ["    - {metric}".format(metric=metric) for metric in sorted(
                                requested_external - set(available_external_metrics))]
                            ))
            )
            sys.exit(1)

        score_keys += __scores

        score_keys = ["tid", "alias", "parent", "score"] + sorted(score_keys)
        # Define mandatory output files
        locus_metrics_file = open(re.sub("$", ".metrics.tsv", re.sub(
            ".gff.?$", "", self.locus_out)), "w")
        locus_scores_file = open(re.sub("$", ".scores.tsv", re.sub(
            ".gff.?$", "", self.locus_out)), "w")

        metrics = Superlocus.available_metrics[5:]
        # Only report *used* external metrics, for performance
        if self.configuration.pick.output_format.report_all_external_metrics is True:
            metrics.extend(available_external_metrics)
        else:
            metrics.extend(requested_external)
        metrics.extend(attribute_metrics)
        metrics = Superlocus.available_metrics[:5] + sorted(metrics)
        session.close()
        engine.dispose()

        locus_metrics = csv.DictWriter(
            locus_metrics_file,
            metrics,
            extrasaction="ignore",
            delimiter="\t")
        locus_metrics.writeheader()
        locus_scores = csv.DictWriter(locus_scores_file, score_keys, delimiter="\t", extrasaction="ignore",)
        locus_scores.writeheader()

        locus_metrics.handle = locus_metrics_file
        locus_metrics.flush = locus_metrics.handle.flush
        locus_metrics.close = locus_metrics.handle.close
        locus_metrics.name = locus_metrics.handle.name
        locus_scores.handle = locus_scores_file
        locus_scores.flush = locus_scores.handle.flush
        locus_scores.close = locus_scores.handle.close
        locus_scores.name = locus_scores.handle.name

        locus_out = open(self.locus_out, 'w')
        sub_files, mono_files = self.__print_gff_headers(locus_out, score_keys, metrics)

        return ((locus_metrics,
                 locus_scores,
                 locus_out),
                sub_files, mono_files)

    # This method has many local variables, but most (9!) are
    # actually file handlers. I cannot trim them down for now.
    # pylint: disable=too-many-locals

    def _submit_locus(self, slocus, counter, engine=None):
        """
        Private method to submit / start the analysis of a superlocus in input.
        :param slocus: the locus to analyse.
        :param data_dict: the preloaded data in memory
        :param engine: connection engine
        :return: job object / None
        """

        if slocus is not None:
            self.main_logger.debug("Submitting %s", slocus.id)
        else:
            return []
            
        self.logger.debug("Loading data for %s", slocus.id)
        slocus.logger = self.logger
        return analyse_locus(slocus=slocus,
                             counter=counter,
                             configuration=self.configuration,
                             logging_queue=self.logging_queue,
                             engine=engine)

    def __unsorted_interrupt(self, row, current_transcript):
        """
        Private method that brings the program to a screeching halt
         if the GTF/GFF is not properly sorted.
        :param row:
        :param current_transcript:
        :return:
        """

        # self.result_dict[counter] = [_]
        self.printer_queue.put_nowait(("EXIT", None))
        # self.printer_queue.put(("EXIT", counter + 1))
        current = "\t".join([str(x) for x in row])
        previous = "\t".join([str(x) for x in [current_transcript.chrom,
                                               current_transcript.start,
                                               current_transcript.end]])

        error_msg = """Unsorted input file, the results will not be correct.
            Please provide a properly sorted input. Error:
            {0}
            {1}""".format(current,
                          previous)
        self.logger.critical(error_msg)
        error_msg = "CRITICAL - {0}".format(" ".join(
            [l.strip() for l in error_msg.split("\n")]))
        raise UnsortedInput(error_msg)

    def __test_sortedness(self, row_coords, coords):
        """
        Private method to test whether a row and the current transcript are actually in the expected
        sorted order.
        :param row_coords:
        :param coords:
        :return:
        """
        test = True
        if coords is None:
            return test

        if not hasattr(coords, "chrom"):
            chrom, start, end = coords
        else:
            chrom, start, end = coords.chrom, coords.start, coords.end

        if not hasattr(row_coords, "chrom"):
            row_chrom, row_start, row_end = coords
        else:
            row_chrom, row_start, row_end = row_coords.chrom, row_coords.start, row_coords.end

        if any(_ is None for _ in (chrom, start, end)):
            return True
        if chrom > row_chrom:
            test = False
        elif (chrom == row_chrom and
              start > row_start):
            test = False
        elif (chrom == row_chrom and
              start == row_start and
              end > row_end):
            test = False

        if test is False:
            self.__unsorted_interrupt(row_coords, coords)

    @staticmethod
    def add_to_index(transcripts: dict,
                     counter: int,
                     locus_queue,
                     mapper: dict,):

        """Method to create the simple indexed database for features."""

        chroms = set([_["chrom"] for _ in transcripts.values()])
        if len(chroms) > 1:
            raise AssertionError(chroms)

        transcripts = msgpack.dumps([_ for _ in transcripts.values()])
        # cursor.execute("INSERT INTO transcripts VALUES (?, ?)", (counter, transcripts))
        # max_submit = 1000
        chrom = chroms.pop()
        if "done" not in mapper:
            mapper["done"] = set()
        if "submit" not in mapper:
            mapper["submit"] = set()

        if chrom not in mapper:
            mapper[chrom] = dict()
            mapper[chrom]["submit"] = set()
            mapper[chrom]["done"] = set()

        mapper[chrom]["submit"].add(counter)
        mapper["submit"].add(counter)
        mapper[counter] = chrom
        locus_queue.put((counter, transcripts))

        return mapper

    def __submit_multi_threading(self):

        """
        Method to execute Mikado pick in multi threaded mode.
        :return:
        """

        intron_range = self.configuration.pick.run_options.intron_range
        self.logger.debug("Intron range: %s", intron_range)

        locus_queue = self.manager.JoinableQueue(-1)
        status_queue = self.manager.JoinableQueue(-1)

        handles = list(self.__get_output_files())
        self.logger.info("Starting Mikado with multiple processes")

        self.logger.debug("Creating the worker processes")

        working_processes = [LociProcesser(msgpack.dumps(dataclasses.asdict(self.configuration)),
                                           locus_queue,
                                           self.logging_queue,
                                           status_queue,
                                           _)
                             for _ in range(1, self.procs+1)]
        # Start all processes
        [_.start() for _ in working_processes]

        self.logger.debug("Started all %d workers", self.procs)
        # No sense in keeping this data available on the main thread now

        try:
            mapper = self.__parse_multithreaded(locus_queue)
        except UnsortedInput:
            [_.terminate() for _ in working_processes]
            raise

        self.logger.debug("Joining children processes")

        percs = percentile(range(1, max(mapper["submit"]) + 1),
                           range(10, 101, 10))
        curr_perc = 0
        total = len(mapper["submit"])
        mapper["results"] = dict()

        while mapper["done"] != mapper["submit"]:
            results = status_queue.get()
            if results[0] == "EXIT":
                self.logger.error("We have not retrieved all loci!")
                raise ValueError
            results = msgpack.loads(zlib.decompress(results))
            self.logger.info(f"Length of results to analyse: {len(results)}")
            for result in results:
                try:
                    counter, chrom, num_genes, loci, subloci, monoloci = result
                except (InvalidTranscript, InvalidCDS, TypeError, RuntimeError, ValueError, AssertionError) as exc:
                    self.logger.critical(
                        f"Mikado has encountered a fatal crash. Invalid result:\n{result}\n"
                        "Please inspect the logs and report the bug at "
                        "https://github.com/EI-CoreBioinformatics/mikado/issues. Exception:\n"
                        f"{exc}")
                    [_.kill() for _ in working_processes]
                    sys.exit(1)
                mapper["done"].add(counter)
                if counter in mapper["results"]:
                    self.logger.fatal("%d double index found!", counter)
                    raise KeyError

                mapper["results"][counter] = (chrom, num_genes, loci, subloci, monoloci)
                if len(mapper["done"]) > percs[curr_perc]:
                    curr_perc += 1
                    while len(mapper["done"]) > percs[curr_perc]:
                        curr_perc += 1
                    real_perc = round(len(mapper["done"]) * 100 / total)
                    self.logger.info("Done %s%% of loci (%s out of %s)", real_perc,
                                     len(mapper["done"]), total)
                chrom = mapper[counter]
                mapper[chrom]["done"].add(counter)
                if mapper[chrom]["done"] == mapper[chrom]["submit"]:
                    self.logger.info("Finished with chromosome %s", chrom)
            status_queue.task_done()

        self.logger.info("Joining children processes")
        [_.join() for _ in working_processes]
        self.logger.info("Joined children processes; starting to merge partial files")

        # Merge loci
        merge_loci(mapper,
                   handles,
                   total=total,
                   configuration=self.configuration,
                   logger=self.logger,
                   source=self.configuration.pick.output_format.source)

        self.logger.info("Finished merging partial files")
        for group in handles:
            [_.close() for _ in group if _]
        return

    def __check_max_intron(self, current, invalids, row, max_intron):
        previous = None
        if hasattr(row, "feature"):
            chrom, feature, start, end, phase, tid = (row.chrom, row.feature, row.start,
                                                      row.end, row.phase, row.transcript)
        else:
            # return line, chrom, start, end, tid, is_transcript
            _, chrom, feature, start, end, phase, tid, _ = row

        for exon in reversed(current["transcripts"][tid]["exon_lines"]):
            if exon[2] == feature:
                previous = exon
                break

        current["transcripts"][tid]["exon_lines"].append((start, end, feature, phase))
        if previous:
            # I have to compare like with like.
            intron_length = (start - 1) - (previous[1] + 1) + 1
            if intron_length >= max_intron:
                self.logger.warning(
                    "%s has an intron (%s) greater than the maximum allowed (%s). Ignoring it.",
                    tid, intron_length, max_intron
                )
                del current["transcripts"][tid]
                invalids.add(tid)
                if current["transcripts"]:
                    current["start"] = min(
                        current["transcripts"][trans]["start"]
                        for trans in current["transcripts"]
                    )

                    current["end"] = max(
                        current["transcripts"][trans]["end"]
                        for trans in current["transcripts"]
                    )

                else:
                    current["start"], current["end"] = None, None
        return current, invalids

    transcript_gtf_pattern = re.compile(r"""transcript_id "([^"]*)\"""")

    def _parse_gtf_line(self, line):

        if not line or line[0] == "#":
            return None
        fields = line.split("\t")
        if len(fields) != 9:
            return None

        try:
            start, end = int(fields[3]), int(fields[4])
        except (ValueError, SystemError, TypeError):
            return None
        chrom = fields[0]
        is_exon = (GtfLine.exon_pattern.search(fields[2]) is not None)
        is_transcript = False
        if not is_exon:
            is_transcript = (GtfLine.transcript_pattern.search(fields[2]) is not None)

        if not (is_exon or is_transcript):
            return None

        tid = self.transcript_gtf_pattern.search(fields[-1])
        if tid is None:
            raise InvalidConfiguration("Corrupt input GTF file, offending line:\n{}".format(line))
        tid = tid.groups()[0]
        if fields[7] in (None, ".", "?"):
            phase = None
        else:
            try:
                phase = int(fields[7])
            except (SystemError, TypeError, ValueError):
                return None
        return line, chrom, fields[2], start, end, phase, tid, is_transcript

    def __parse_multithreaded(self, locus_queue):
        counter = 0
        invalids = set()
        flank = self.configuration.pick.clustering.flank
        mapper = dict()

        with self.define_input(multithreading=True) as input_annotation:
            current = {"chrom": None, "start": None, "end": None, "transcripts": dict()}
            max_intron = self.configuration.prepare.max_intron_length
            for row in input_annotation:
                row = self._parse_gtf_line(row)
                if row is None:  # Header
                    continue
                line, chrom, feature, start, end, phase, tid, is_transcript = row
                if self.__regions and chrom not in self.__regions:
                    continue

                if is_transcript is False:
                    if tid in invalids:
                        continue
                    elif tid not in current["transcripts"]:
                        self.logger.fatal("Transcript %s is invalid", tid)
                        raise UnsortedInput
                        # invalids.add(tid)
                    else:
                        # Check max intron length. We presume that exons are sorted correctly.
                        current, invalids = self.__check_max_intron(current, invalids,
                                                                    row, max_intron=max_intron)
                elif is_transcript is True:
                    self.__test_sortedness((chrom, start, end),
                                           (current["chrom"], current["start"], current["end"]))
                    if self.__regions:
                        if not [_ for _ in self.__regions[chrom].find(start, end) if _.start <= start and end <= _.end]:
                            invalids.add(tid)
                            continue

                    if current["chrom"] != chrom or not current["start"]:
                        if current["chrom"] != chrom:
                            if current["chrom"] is not None and current["chrom"] != chrom:
                                self.logger.debug("Finished chromosome %s", current["chrom"])
                                if len(current["transcripts"]) > 0:
                                    self.logger.debug("Submitting locus # %d (%s), with transcripts:\n%s",
                                                      counter, "{}:{}-{}".format(current["chrom"],
                                                                                 current["start"], current["end"]),
                                                      ",".join(list(current["transcripts"].keys())))
                                    counter += 1
                                    mapper = self.add_to_index(current["transcripts"], counter, locus_queue, mapper)
                            self.logger.debug("Starting chromosome %s", chrom)

                        current["chrom"], current["start"], current["end"] = chrom, start, end
                        current["transcripts"] = dict()
                        current["transcripts"][tid] = dict()
                        current["transcripts"][tid]["chrom"] = chrom
                        current["transcripts"][tid]["start"] = start
                        current["transcripts"][tid]["end"] = end
                        current["transcripts"][tid]["definition"] = line
                        current["transcripts"][tid]["exon_lines"] = []
                    else:
                        if Superlocus.overlap((current["end"], current["start"]),
                                              (start, end), flank=flank) > 0:
                            # Add to the locus!
                            current["start"] = min(current["start"], start)
                            current["end"] = max(current["end"], end)
                        elif len(current["transcripts"]) > 0:
                            counter += 1
                            self.logger.debug("Submitting locus # %d (%s), with transcripts:\n%s",
                                              counter, "{}:{}-{}".format(current["chrom"],
                                                                         current["start"], current["end"]),
                                              ",".join(list(current["transcripts"].keys())))
                            mapper = self.add_to_index(current["transcripts"], counter, locus_queue, mapper)
                            current["start"], current["end"] = start, end
                            current["transcripts"] = dict()

                        current["transcripts"][tid] = dict()
                        current["transcripts"][tid]["chrom"] = chrom
                        current["transcripts"][tid]["definition"] = line
                        current["transcripts"][tid]["exon_lines"] = []
                        current["transcripts"][tid]["start"] = start
                        current["transcripts"][tid]["end"] = end

        if current["start"] is not None:
            if len(current["transcripts"]) > 0:
                counter += 1
                self.logger.debug("Submitting locus # %d (%s), with transcripts:\n%s",
                                  counter, "{}:{}-{}".format(current["chrom"],
                                                             current["start"], current["end"]),
                                  ",".join(list(current["transcripts"].keys())))
                mapper = self.add_to_index(current["transcripts"], counter, locus_queue, mapper)
            self.logger.debug("Finished chromosome %s", current["chrom"])

        locus_queue.put(("EXIT", None))

        return mapper

    def __submit_single_threaded(self):

        """
        Method to execute Mikado pick in single threaded mode.
        :return:
        """

        current_locus = None
        current_transcript = None

        handler = logging_handlers.QueueHandler(self.logging_queue)
        logger = logging.getLogger("queue_listener")
        logger.propagate = False
        logger.addHandler(handler)
        logger.setLevel(self.configuration.log_settings.log_level)
        logger.debug("Begun single-threaded run")

        intron_range = self.configuration.pick.run_options.intron_range
        logger.debug("Intron range: %s", intron_range)

        handles = self.__get_output_files()

        locus_printer = functools.partial(print_locus,
                                          handles=handles,
                                          logger=logger,
                                          configuration=self.configuration)

        # last_printed = -1
        curr_chrom = None
        gene_counter = 0

        self.engine = dbutils.connect(configuration=self.configuration, logger=self.logger)

        submit_locus = functools.partial(self._submit_locus, **{"engine": self.engine})

        counter = -1
        invalid = False
        max_intron = self.configuration.prepare.max_intron_length
        skip_transcript = False
        with self.define_input() as input_annotation:
            for row in input_annotation:
                if row.is_exon is True and not invalid and not skip_transcript:
                    try:
                        current_transcript.add_exon(row)
                    except InvalidTranscript as exc:
                        self.logger.error("Transcript %s is invalid;\n%s",
                                          current_transcript.id,
                                          exc)
                        invalid = True
                elif row.is_transcript is True:
                    self.__test_sortedness(row, current_transcript)
                    current_locus, counter, gene_counter, curr_chrom = self.__check_transcript(
                        current_transcript, current_locus, counter, max_intron,
                        gene_counter, curr_chrom, locus_printer, submit_locus)
                    invalid = False
                    if self.__regions:
                        if row.chrom not in self.__regions:
                            skip_transcript = True
                            current_transcript = None
                        elif not [_ for _ in self.__regions[row.chrom].find(row.start, row.end)
                                  if _.start <= row.start and row.end <= _.end]:
                            skip_transcript = True
                            current_transcript = None
                        else:
                            current_transcript = Transcript(row, intron_range=intron_range, logger=logger)
                            skip_transcript = False
                    else:
                        current_transcript = Transcript(row, intron_range=intron_range, logger=logger)
                        skip_transcript = False
                    if not skip_transcript and (
                        (
                            current_transcript is None
                            or row.chrom != current_transcript.chrom
                        )
                    ):
                        if current_transcript is not None:
                            self.logger.info("Finished chromosome %s",
                                             current_transcript.chrom)
                        self.logger.info("Starting chromosome %s", row.chrom)

        self.logger.info("Finished chromosome %s", current_locus.chrom)
        current_locus, counter, gene_counter, curr_chrom = self.__check_transcript(
            current_transcript, current_locus, counter, max_intron,
            gene_counter, curr_chrom, locus_printer, submit_locus)
        counter += 1
        self.logger.debug("Analysing locus # %d", counter)
        for stranded_locus in submit_locus(current_locus, counter):
            if stranded_locus.chrom != curr_chrom:
                curr_chrom = stranded_locus.chrom
                gene_counter = 0
            gene_counter = locus_printer(stranded_locus, gene_counter)
        # submit_locus(current_locus, counter)
        for group in handles:
            [_.close() for _ in group if _]
        logger.info("Final number of superloci: %d", counter)

    def __check_transcript(self, current_transcript, current_locus, counter, max_intron,
                           gene_counter, curr_chrom, locus_printer, submit_locus):

        if current_transcript is not None:
            current_transcript.finalize()
            if current_transcript.max_intron_length > max_intron:
                self.logger.warning(
                    "%s has an intron (%s) which is greater than the maximum allowed (%s). Removing it.",
                    current_transcript.id, current_transcript.max_intron_length, max_intron)
            elif current_locus and Superlocus.in_locus(
                            current_locus, current_transcript,
                            flank=self.configuration.pick.clustering.flank) is True:
                current_locus.add_transcript_to_locus(current_transcript, check_in_locus=False)
            else:
                counter += 1
                self.logger.debug("Analysing locus # %d", counter)
                try:
                    for stranded_locus in submit_locus(current_locus, counter):
                        if stranded_locus.chrom != curr_chrom:
                            curr_chrom = stranded_locus.chrom
                            gene_counter = 0
                        gene_counter = locus_printer(stranded_locus, gene_counter)
                except KeyboardInterrupt:
                    raise
                except Exception as exc:
                    self.logger.exception(
                        "Superlocus %s failed with exception: %s",
                        None if current_locus is None else current_locus.id, exc)

                current_locus = Superlocus(current_transcript, stranded=False, configuration=self.configuration,
                                           source=self.configuration.pick.output_format.source)

        return current_locus, counter, gene_counter, curr_chrom

    def _parse_and_submit_input(self):

        """
        This method does the parsing of the input and submission of the loci to the
        _submit_locus method.
        :return: jobs (the list of all jobs already submitted)
        """

        single_thread = (self.configuration.pick.run_options.single_thread or self.procs == 1
                         or self.configuration.log_settings.log_level == "DEBUG")
        
        if single_thread is False:
            self.__submit_multi_threading()
        else:
            self.__submit_single_threaded()
        return

    def __cleanup(self):
        self.log_writer.stop()
        if self.queue_pool is not None:
            self.queue_pool.dispose()
        self.manager.shutdown()

    def __call__(self):

        """This method will activate the class and start the analysis of the input file."""

        # NOTE: Pool, Process and Manager must NOT become instance attributes!
        # Otherwise it will raise all sorts of mistakes

        self.logger.debug("Source: %s",
                          self.configuration.pick.output_format.source)
        if self.configuration.db_settings.dbtype == "sqlite":
            self.queue_pool = sqlalchemy.pool.QueuePool(
                self.db_connection,
                pool_size=self.procs,
                max_overflow=0)

        try:
            self._parse_and_submit_input()
        except UnsortedInput as _:
            self.logger.error(
                "The input files were not properly sorted! Please run prepare and retry.")
        except Exception:
            self.__cleanup()
            raise

        self.__cleanup()
        self.main_logger.info("Finished analysis of %s", self.input_file)

        sys.exit(0)
# pylint: enable=too-many-instance-attributes



#!/usr/bin/env python3

"""
Collection of utilities that are useful for managing Mikado-related files,
including e.g. a statistical calculator for GFF/GTFs, an awk-like utility for
GTFs, and a grep-like utility for annotation files.
"""


from . import awk_gtf
from . import metrics
from . import stats
from . import trim
from . import grep
from . import convert
from . import class_codes
from . import collapse_compare_stats
import argparse

__author__ = 'Luca Venturini'


def util_parser():
    """
    Function to return all available utility parsers.
    :rtype: argparse.ArgumentParser
    """

    desc = """Collection of utilities for managing GTF/GFF files."""
    parser = argparse.ArgumentParser(prog="util",
                                     description=desc)
    utils = parser.add_subparsers()

    utils.add_parser("awk_gtf",
                     description="Script to retrieve specific feature slices from a GTF file.")
    utils.choices["awk_gtf"] = awk_gtf.awk_parser()
    utils.choices["awk_gtf"].prog = "mikado util awk_gtf"
    utils.choices["awk_gtf"].description = "Script to retrieve specific feature slices from a GTF file."

    utils.add_parser("class_codes",
                     description="Script to print out the class codes.")
    utils.choices["class_codes"] = class_codes.code_parser()
    utils.choices["class_codes"].prog = "mikado util class_codes"
    utils.choices["class_codes"].description = "Script to print out the class codes."

    utils.add_parser("collect_compare",
                     description="Script to collect and collapse multiple `mikado compare` summary statistics.")
    utils.choices["collect_compare"] = collapse_compare_stats.parser()
    utils.choices["collect_compare"].prog = "mikado util collect_compare"
    utils.choices["collect_compare"].description = "Script to collect and collapse multiple `mikado compare` "\
                                                   "summary statistics."

    utils.add_parser("convert",
                     description="Script to do BAM -> GTF <-> GFF3 -> BED12 conversions.")
    utils.choices["convert"] = convert.convert_parser()
    utils.choices["convert"].prog = "mikado util convert"
    utils.choices["convert"].description = "Script to do BAM -> GTF <-> GFF3 -> BED12 conversions."

    utils.add_parser("grep",
                     description="Script to extract specific models from GFF/GTF files.")
    utils.choices["grep"] = grep.grep_parser()
    utils.choices["grep"].prog = "mikado util grep"
    utils.choices["grep"].description = "Script to extract specific models from GFF/GTF files."

    utils.add_parser("metrics",
                     description="Simple script to obtain the documentation on the transcript metrics.")
    utils.choices["metrics"] = metrics.metric_parser()
    utils.choices["metrics"].prog = "mikado util metrics"
    utils.choices["metrics"].description = "Simple script to obtain the documentation on \
the transcript metrics."

    utils.add_parser("stats", description="""GFF/GTF statistics script.
    It will compute median/average length of RNAs, exons, CDS features, etc.""")
    utils.choices["stats"] = stats.stats_parser()
    utils.choices["stats"].prog = "mikado util stats"
    utils.choices["stats"].description = "GFF/GTF statistics script. \
It will compute median/average length of RNAs, exons, CDS features, etc."

    utils.add_parser("trim",
                     description="Script to remove up to N bps from terminal exons in an annotation file.")
    utils.choices["trim"] = trim.trim_parser()
    utils.choices["trim"].prog = "mikado util trim"
    utils.choices["trim"].description = "Script to remove up to N bps from terminal\
exons in an annotation file."

    parser.add_help = True

    return parser



#!/usr/bin/env python3
# coding: utf-8

"""GFF +> BED12 converter for junction files."""

import argparse
import sys

import Mikado.loci
import Mikado.parsers


def main():
    """Main script function"""

    parser = argparse.ArgumentParser("GFF=>BED12 converter")
    parser.add_argument("gff", type=argparse.FileType())
    parser.add_argument("out", nargs="?", type=argparse.FileType("w"), default=sys.stdout)
    args = parser.parse_args()

    transcript = None

    for row in Mikado.parsers.GFF.GFF3(args.gff):
        if row.is_parent and not row.is_transcript:
            continue
        if row.is_exon:
            transcript.add_exon(row)
        elif row.is_transcript:
            if transcript is not None:
                transcript.finalize()
                assert len(transcript.introns) == 1
                introns = sorted(transcript.introns)
                bed12 = Mikado.parsers.bed12.BED12()
                bed12.chrom = transcript.chrom
                bed12.strand = transcript.strand
                bed12.score = transcript.score
                bed12.start = transcript.start
                bed12.end = transcript.end
                bed12.thick_start = introns[0][0]
                bed12.thick_end = introns[0][1]
                bed12.name = transcript.id
                bed12.rgb = "255,0,0"
                bed12.block_count = 2
                bed12.block_sizes = [(e[1] - e[0]) for e in transcript.exons]
                bed12.block_starts = []
                bed12.block_starts = [0, bed12.block_sizes[0] + introns[0][1] - introns[0][0]]
                print(bed12, file=args.out)

            transcript = Mikado.transcripts.transcript.Transcript(row)

    if transcript is not None:
        transcript.finalize()
        assert len(transcript.introns) == 1
        introns = sorted(transcript.introns)
        bed12 = Mikado.parsers.bed12.BED12()
        bed12.chrom = transcript.chrom
        bed12.strand = transcript.strand
        bed12.start = transcript.start
        bed12.score = transcript.score
        bed12.end = transcript.end
        bed12.thick_start = introns[0][0]
        bed12.thick_end = introns[0][1]
        bed12.name = transcript.id
        bed12.rgb = "255,0,0"
        bed12.block_count = 2
        bed12.block_sizes = [(e[1] - e[0]) for e in transcript.exons]
        bed12.block_starts = []
        bed12.block_starts = [0, bed12.block_sizes[0] + introns[0][1] - introns[0][0]]
        print(bed12, file=args.out)


if __name__ == "__main__":
    main()



# cython: language_level=3

import numpy as np
cimport numpy as np
from libc.stdlib cimport malloc
from libcpp.string cimport string as cstring
cimport cython

DTYPE = np.int_
ctypedef np.int_t DTYPE_t


@cython.cdivision(True)
cdef _analyze_string(char* qseq, char* sseq, char* mid,
                     long query_start, long query_end, long query_length, long qmult):

    cdef long qpos = -1
    cdef long shape = <long> (query_end - query_start) / qmult
    cdef np.ndarray[DTYPE_t, ndim=1] query_array = np.zeros(shape, dtype=DTYPE)
    cdef DTYPE_t[:] query_view = query_array
    cdef Py_ssize_t match_len = len(mid)
    cdef cstring match
    cdef char qchar, schar, midchar
    cdef Py_ssize_t idx
    if shape > query_length:
        raise ValueError((shape, query_length, query_start, query_end, qmult, mid))

    for idx in range(match_len):
        qchar, schar, midchar = qseq[idx], sseq[idx], mid[idx]
        if qchar == b"-":
            if schar == b"*":
                match.push_back(b"*")
            else:
                match.push_back(b"-")
            continue
        else:
            qpos += 1
            if qchar == b"*" or schar == b"*":
                match.push_back(b"*")
            elif qchar == schar:
                match.push_back(b"|")
                query_view[qpos] = 2
            elif midchar == b"+":
                match.push_back(b"+")
                query_view[qpos] = 1
            elif schar == b"-":
                if qchar == b"*":
                    match.push_back(b"*")
                else:
                    match.push_back(b"_")
                continue
            else:
                match.push_back(b"/")

    return query_array, match


cpdef prepare_aln_strings(hsp, bint off_by_one=0, long qmultiplier=1):

    """This private method calculates the identical positions, the positives, and a re-factored match line
    starting from the HSP.
    :type hsp: Bio.SearchIO.HSP
    """

    cdef bytes qseq = <bytes> str(hsp.query.seq).encode()
    cdef bytes sseq = <bytes> str(hsp.hit.seq).encode()
    cdef bytes mid = <bytes> str(hsp.aln_annotation["similarity"]).encode()
    cdef long match_len = len(mid)

    lett_array = np.array([list(str(hsp.query.seq)),
                           list(str(hsp.aln_annotation["similarity"])),
                           list(str(hsp.hit.seq))], dtype=np.str_)

    if lett_array.shape[0] == 0 or len(lett_array.shape) == 1 or lett_array.shape[1] == 0:  # Empty array!
        raise ValueError("Empty array of matches! {}".format("\n".join(
            [hsp.query.seq, hsp.hit.seq, hsp.aln_annotation["similarity"]])))

    cdef np.ndarray[DTYPE_t, ndim=1] query_array
    cdef long query_start = hsp.query_start
    cdef long query_end = hsp.query_end - off_by_one
    cdef long query_length = len(qseq)
    cdef np.ndarray[DTYPE_t, ndim=2] summer = np.array([[_] for _ in range(qmultiplier)])
    query_array, match = _analyze_string(qseq, sseq, mid, query_start, query_end, query_length, qmultiplier)
    cdef np.ndarray[DTYPE_t, ndim=1] _id_catcher = np.where(query_array >= 2)[0]
    identical_positions = ((_id_catcher * qmultiplier) + summer).flatten()
    cdef np.ndarray[DTYPE_t, ndim=1] _pos_catcher = np.where(query_array >= 1)[0]
    positives = ((_pos_catcher * qmultiplier) + summer).flatten()
    if hsp.query_frame > 0:
        identical_positions = identical_positions + hsp.query_start
        positives = positives + hsp.query_start
    else:
        identical_positions = hsp.query_end - identical_positions - 1
        positives = hsp.query_end - positives - 1

    # identical_positions = set(identical_positions)
    # positives = set(positives)
    identical_positions.sort()
    positives.sort()
    match = <bytes>match
    match = match.decode()

    return match, identical_positions, positives



from .hit import Hit
from .hsp import Hsp
import sqlalchemy.exc


def load_into_db(self, hits, hsps, force=False, raw=False):
    """
    :param hits:
    :param hsps:

    :param force: boolean flag. If set, data will be loaded no matter what.
    To be used at the end of the serialisation to load the final batch of data.
    :type force: bool

    :return:
    """

    # self.logger.debug("Checking whether to load %d hits and %d hsps", len(hits), len(hsps))

    tot_objects = len(hits) + len(hsps)
    if len(hits) == 0:
        self.logger.debug("No hits to serialise. Exiting")
        return hits, hsps

    if tot_objects >= self.maxobjects or force:
        # Bulk load
        self.logger.debug("Loading %d BLAST objects into database", tot_objects)

        if not hasattr(self, "hit_i_string"):
            self.hit_i_string = str(Hit.__table__.insert(bind=self.engine).compile())
            self.hsp_i_string = str(Hsp.__table__.insert(bind=self.engine).compile())

        try:
            # pylint: disable=no-member
            # self.session.begin(subtransactions=True)
            if hasattr(self, "lock") and self.lock is not None:
                self.lock.acquire()
            if raw is True:
                self.engine.execute(self.hit_i_string, hits)
                self.engine.execute(self.hsp_i_string, hsps)
            else:
                self.engine.execute(Hit.__table__.insert(), hits)
                self.engine.execute(Hsp.__table__.insert(), hsps)
            if hasattr(self, "lock") and self.lock is not None:
                self.lock.release()
            # pylint: enable=no-member
        except sqlalchemy.exc.IntegrityError as err:
            self.logger.critical("Failed to serialise BLAST!")
            self.logger.exception(err)
            raise err
        self.logger.debug("Loaded %d BLAST objects into database", tot_objects)
        hits, hsps = [], []

    if force is True: # Final push
        if hasattr(self, "lock") and self.lock is not None:
            self.lock.acquire()
        self.session.commit()
        if hasattr(self, "lock") and self.lock is not None:
            self.lock.release()

    return hits, hsps



import asyncio
import itertools
from Mikado._transcripts.scoring_configuration import MinMaxScore, SizeFilter
from Mikado.configuration.configurator import load_and_validate_config
from Mikado.loci import Superlocus
from Mikado.parsers.bed12 import BED12
from Mikado.serializers.blast_serializer import Target, Hit, Hsp
from Mikado.serializers.external import External, ExternalSource
from Mikado.serializers.blast_serializer.query import Query
from Mikado.serializers.orf import Orf
from Mikado.serializers.junction import Chrom, Junction
from Mikado.transcripts import Transcript
from Mikado.utilities.dbutils import DBBASE as db
import unittest
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine


class AsyncExternalTest(unittest.TestCase):

    def test_get_external(self):
        checked_conf = load_and_validate_config(None).copy()
        checked_conf.pick.output_format.report_all_external_metrics = True
        transcript = Transcript()
        transcript.chrom = "15"
        transcript.source = "protein_coding"
        transcript.start = 47631264
        transcript.end = 48051999

        exons = [(47631264, 47631416),
                 (47704590, 47704669),
                 (47762671, 47762742),
                 (47893062, 47893093),
                 (47895572, 47895655),
                 (48051942, 48051999)]

        transcript.strand = "+"
        transcript.add_exons(exons)
        transcript.id = "ENST00000560636"
        transcript.parent = "ENSG00000137872"
        transcript2 = transcript.copy()
        transcript2.id = "ENST00000560637"
        checked_conf.scoring.scoring["attributes.tpm"] = MinMaxScore.Schema().load(
            {"rescaling": "max", "default": 0, "rtype": "float", 'multiplier': 4,
             'use_raw': True, 'percentage': True})
        transcript.attributes["tpm"] = 10

        int_source = ExternalSource('int', 'int', 0)
        float_source = ExternalSource('float', 'float', 0)
        bool_source = ExternalSource('bool', 'bool', 0)

        raw_int_source = ExternalSource('raw_int', 'int', 1)
        raw_float_source = ExternalSource('raw_float', 'float', 1)
        raw_bool_source = ExternalSource('raw_bool', 'bool', 1)

        int_score = External(1, 1, 10)
        float_score = External(1, 2, 10.0)
        bool_score = External(1, 3, int(False))  # We cast as int here following external.py serialize function

        raw_int_score = External(1, 4, 8)
        raw_float_score = External(1, 5, 8.0)
        raw_bool_score = External(1, 6, int(True))  # We cast as int here following external.py serialize function

        query = Query(transcript.id, transcript.cdna_length)
        query2 = Query(transcript2.id, transcript2.cdna_length)

        engine = create_engine("sqlite:///:memory:")
        db.metadata.create_all(engine)
        SessionMaker = sessionmaker(bind=engine)
        session = SessionMaker()
        session.add_all([int_source, float_source, bool_source, raw_int_source, raw_float_source, raw_bool_source])
        session.add_all([query, query2])
        session.add_all([int_score, float_score, bool_score, raw_int_score, raw_float_score, raw_bool_score])
        session.commit()
        sup = Superlocus(transcript, configuration=checked_conf)
        sup.session = session
        tid = transcript.id
        self.assertIn(tid, sup.transcripts)
        from collections import namedtuple
        qobj = {1: namedtuple('t', field_names=('query_name'))}
        qobj[1].query_name = 'ENST00000560636'
        external = asyncio.run(sup.get_external(qobj, [1]))

        self.assertEqual(external, {
            'ENST00000560636': {
                   'int': (10, False),
                   'float': (10.0, False),
                   'bool': (False, False),
                   'raw_int': (8, True),
                   'raw_float': (8.0, True),
                   'raw_bool': (True, True)
            }
        })

        sup.configuration.pick.output_format.report_all_external_metrics = False
        external = asyncio.run(sup.get_external(qobj, [1]))
        self.assertEqual(len(external), 0)
        # These are meaningless it's just to verify we are loading *only* these metrics.
        # We should *NOT* have 'float' as it is not present in any section.
        sup.configuration.scoring.scoring["external.int"] = MinMaxScore(rescaling="max", filter=None)
        sup.configuration.scoring.requirements.parameters["external.raw_float"] = SizeFilter(operator="gt",
                                                                                             value=100)
        sup.configuration.scoring.cds_requirements.parameters["external.raw_int"] = SizeFilter(operator="lt",
                                                                                                 value=1)
        sup.configuration.scoring.as_requirements.parameters["external.raw_bool"] = SizeFilter(operator="lt",
                                                                                                 value=1)
        sup.configuration.scoring.not_fragmentary.parameters["external.bool"] = SizeFilter(operator="ne",
                                                                                           value=False)
        external = asyncio.run(sup.get_external(qobj, [1]))
        self.assertEqual(external, {
            'ENST00000560636': {
                'int': (10, False),
                'raw_float': (8.0, True),
                'bool': (False, False),
                'raw_int': (8, True),
                'raw_bool': (True, True)
            }
        })


class AsyncJunctionTest(unittest.TestCase):

    def test_retrieval(self):
        engine = create_engine("sqlite:///:memory:")
        db.metadata.create_all(engine)
        SessionMaker = sessionmaker(bind=engine)
        session = SessionMaker()

        transcript = Transcript(accept_undefined_multi=True)
        transcript.chrom = "15"
        transcript.source = "protein_coding"
        transcript.start = 47631264
        transcript.end = 48051999

        exons = [(47631264, 47631416),
                 (47704590, 47704669),
                 (47762671, 47762742),
                 (47893062, 47893093),
                 (47895572, 47895655),
                 (48051942, 48051999)]

        transcript.strand = "+"
        transcript.add_exons(exons)
        transcript.id = "ENST00000560636"
        transcript.parent = "ENSG00000137872"
        transcript2 = transcript.copy()
        transcript2.id = "ENST00000560637"

        chrom_one = Chrom("1", 10**8)
        chrom_fifteen = Chrom("15", 5 * 10 ** 8)
        session.add_all([chrom_one, chrom_fifteen])
        session.commit()
        # junction_start, junction_end, name, strand, score, chrom_id)
        # This junction is on a different chrom
        junction_chrom_one = Junction(47704669 + 1, 47762671 - 1, "chrom_one", "+", 10, chrom_one.chrom_id)
        # This junction is too far away
        outside_chrom_15 = Junction(47704669 - 10 ** 6 + 1, 47762671 - 10 ** 6 - 1, "chrom_15_outside", "+", 10,
                                    chrom_fifteen.chrom_id)
        # This junction is in the right place but wrong strand
        wrong_strand_chrom_15 = Junction(47704669 + 1, 47762671 - 1, "chrom_15_wrong_strand", "-", 10,
                                         chrom_fifteen.chrom_id)
        # This one is correct
        chrom_15_junction = Junction(47704669 + 1, 47762671 - 1, "chrom_15", "+", 10, chrom_fifteen.chrom_id)
        session.add_all([junction_chrom_one, outside_chrom_15, wrong_strand_chrom_15, chrom_15_junction])
        session.commit()

        self.assertEqual(junction_chrom_one.chrom, "1")
        for junc in [outside_chrom_15, wrong_strand_chrom_15, chrom_15_junction]:
            self.assertEqual(junc.chrom, "15")

        for strand, stranded in itertools.product(("+", "-", None), (True, False)):
            transcript.unfinalize()
            transcript.strand = strand
            transcript.finalize()
            sup = Superlocus(transcript, stranded=stranded)
            self.assertTrue((chrom_15_junction.junction_start, chrom_15_junction.end) in
                            sup.introns, (chrom_15_junction, sup.introns))
            sup.session = session
            asyncio.run(sup._load_introns())
            if stranded is True and strand is not None:
                self.assertEqual(sup.locus_verified_introns, {(chrom_15_junction.junction_start,
                                                               chrom_15_junction.junction_end,
                                                               strand)},
                                 (stranded, strand))
            elif stranded is False:
                self.assertEqual(sup.locus_verified_introns, {(chrom_15_junction.junction_start,
                                                               chrom_15_junction.junction_end,
                                                               chrom_15_junction.strand),
                                                              (wrong_strand_chrom_15.junction_start,
                                                               wrong_strand_chrom_15.junction_end,
                                                               wrong_strand_chrom_15.strand)},
                                 (stranded, strand))
            elif stranded is True and strand is None:
                self.assertEqual(sup.locus_verified_introns, set())


class AsyncOrfLoading(unittest.TestCase):

    def test_load_orfs(self):

        transcript_line = 'Chr1\t100\t2000\tID=foo;coding=True;phase=0'\
                          '\t0\t+\t300\t1850\t0\t4\t400,400,400,200\t0,500,1100,1700'
        transcript = Transcript(transcript_line)
        orf = transcript.orfs[0].to_transcriptomic()
        transcript2 = transcript.copy()
        transcript2.unfinalize()
        transcript2.chrom = "Chr2"
        transcript2.id = "foo.2"
        transcript2.finalize()
        other_orf = transcript2.orfs[0].to_transcriptomic()
        engine = create_engine("sqlite:///:memory:")
        db.metadata.create_all(engine)
        SessionMaker = sessionmaker(bind=engine)
        session = SessionMaker()
        query = Query(transcript.id, transcript.cdna_length)
        query2 = Query(transcript2.id, transcript2.cdna_length)
        session.add_all([query, query2])
        session.commit()
        serialized_orf = Orf(orf, query.query_id)
        self.assertEqual(serialized_orf.thick_end, orf.thick_end)
        self.assertEqual(serialized_orf.cds_len, orf.cds_len)
        serialized_other_orf = Orf(other_orf, query2.query_id)
        session.add_all([serialized_orf, serialized_other_orf])
        session.commit()
        sup = Superlocus(transcript)
        sup.session = session
        sup_orfs = asyncio.run(sup.get_orfs([query.query_id]))
        self.assertEqual(len(sup_orfs), 1)
        self.assertIn(transcript.id, sup_orfs)
        self.assertEqual(len(sup_orfs[transcript.id]), 1)
        self.assertIsInstance(sup_orfs[transcript.id][0], BED12, type(sup_orfs[transcript.id][0]))
        self.assertTrue(sup_orfs[transcript.id][0] == orf, "\n" + "\n".join(
            [str(orf), str(sup_orfs[transcript.id][0])]))


# TODO: Create a test for the BLAST hits/hsps

class AsyncBlastTest(unittest.TestCase):
    """Test for the functionality of loading a BLAST hit from a Superlocus object."""



import rapidjson as json
decoder = json.Decoder()
import msgpack
from ..loci import Superlocus
from ..loci import Locus
import sys
import collections
import itertools
import numpy as np
from ._locus_line_creator import _create_locus_lines


def manage_index(index, data, source, configuration):
    index, (chrom, gene_counter, gene_max) = index[0], index[1]
    orig_gene_counter = gene_counter
    batch = []
    chrom, num_genes, stranded_loci, sublocus_dump, monolocus_dump = data[index]

    loci = []
    sublocus_dump = msgpack.loads(sublocus_dump, raw=False)

    sub_length = len(sublocus_dump)

    monolocus_dump = msgpack.loads(monolocus_dump, raw=False)

    mono_length = len(monolocus_dump)

    for pos, stranded_locus_json in enumerate(msgpack.loads(stranded_loci, raw=False)):
        stranded_locus = Superlocus(None, configuration=configuration)
        for locus_string in stranded_locus_json:
            locus_dict = decoder(locus_string)
            locus = Locus(configuration=configuration)
            locus.load_dict(locus_dict, load_configuration=False)
            if locus is not None:
                stranded_locus.add_locus(locus)
        stranded_locus.source = source

        if not stranded_locus.id.endswith(str(sys.maxsize)):
            loci.append(stranded_locus.id)

        minibatch, gene_counter = _create_locus_lines(stranded_locus, gene_counter)
        minibatch = [minibatch]

        if pos < sub_length:
            assert len(sublocus_dump[pos]) == 3, sublocus_dump
            minibatch.append(sublocus_dump[pos])
        else:
            minibatch.append([])

        if pos < mono_length:
            minibatch.append(monolocus_dump[pos])
        else:
            minibatch.append([])

        batch.append(minibatch)

    assert (gene_counter - orig_gene_counter) == gene_max, (orig_gene_counter, gene_counter, gene_max)
    if len(set(loci)) != len(loci):
        seen = set()
        duplicated = []
        for lid in loci:
            if lid in seen:
                duplicated.append(lid)
            else:
                seen.add(lid)
        raise ValueError("Duplicated loci in counter {}! {}".format(index, duplicated))
    batch = [loci, batch]
    # batch = msgpack.dumps(batch)
    return batch


def __create_gene_counters(common_index: dict) -> (dict, int):
    """Function to assign to each counter in the database the correct base and maximum number of genes.
    This allows to parallelise the printing.
    The common index has the following structure:

    d[counter] = (database index, chrom, number of genes in locus)
    """

    chroms = []
    num_genes = []

    for index in range(1, max(common_index.keys()) + 1):
        chrom, n_genes = common_index[index][:2]
        chroms.append(chrom)
        num_genes.append(n_genes)

    chroms = np.array(chroms)
    num_genes = np.array(num_genes)

    gene_counters = dict()
    total_genes = sum(num_genes)

    chrom_tots = collections.defaultdict(list)
    for chrom in np.unique(chroms):
        index = np.where(chroms == chrom)
        totals = num_genes[index]
        cumu = totals.cumsum()
        for counter, former, num in zip(index[0], itertools.chain([0], cumu[:-1]), totals):
            gene_counters[counter + 1] = (former, num)
            if chrom:
                chrom_tots[chrom].extend(list(range(former + 1, former + num + 1)))

    tot_found = 0
    for chrom in chrom_tots:
        if len(set(chrom_tots[chrom])) != len(chrom_tots[chrom]):
            seen = set()
            duplicated = set()
            for num in chrom_tots[chrom]:
                if num in seen:
                    duplicated.add(num)
                else:
                    seen.add(num)
            raise AssertionError((chrom,
                                  len(set(chrom_tots[chrom])),
                                  len(chrom_tots[chrom]), max(chrom_tots[chrom]),
                                  duplicated,
                                  chrom_tots[chrom]))
        if len(chrom_tots[chrom]) > 0:
            assert len(list(range(1, chrom_tots[chrom][-1] + 1))) == len(chrom_tots[chrom])
            tot_found += chrom_tots[chrom][-1]

    assert tot_found == total_genes, (tot_found, total_genes)

    assert min(common_index) == 1

    new_common = dict()
    for key in common_index:
        # chrom, former id, new id
        new_common[key] = (common_index[key][0], gene_counters[key][0], gene_counters[key][1])
    return new_common, total_genes



# coding: utf-8

"""
This module is necessary to serialise external data.
The initial model is very simple - each external data
will have a tag, and internally the files will be TAB-delimited
2-column files, one for the TID and one for the score.
"""

import os
import pyfaidx
from sqlalchemy import Column, String, Integer, ForeignKey, Boolean, inspect
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql.schema import PrimaryKeyConstraint, Index
from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm.session import Session
from ..utilities.dbutils import DBBASE, connect
from .blast_serializer.query import Query
from ..utilities.log_utils import check_logger, create_default_logger
import numbers
import pandas as pd
import numpy as np
import re
from collections import Counter
import sqlalchemy.exc


class ExternalSource(DBBASE):

    __tablename__ = "external_sources"

    source_id = Column(Integer, primary_key=True)
    source = Column(String, unique=True)
    rtype = Column(String, unique=False)
    valid_raw = Column(Boolean)
    __table_args__ = ((Index("external_source_idx", "source_id"),))

    def __init__(self, source, rtype, valid_raw):

        self.source = source
        if valid_raw not in (True, False):
            raise ValueError("\"Valid raw\" flags must be boolean!")
        if np.dtype("bool") == rtype:
            rtype = "bool"
        elif np.dtype("int") == rtype:
            rtype = "int"
        elif np.dtype("float") == rtype:
            rtype = "float"
        elif np.dtype("complex") == rtype:
            rtype = "complex"
        else:
            raise ValueError("Invalid source rtype for {}: {}".format(source, rtype))

        self.rtype = rtype
        self.valid_raw = valid_raw


class External(DBBASE):

    """This class serialises transcript ids, source of the score, and score itself."""

    __tablename__ = "external"

    query_id = Column(Integer, ForeignKey(Query.query_id), unique=False)
    source_id = Column(Integer, ForeignKey(ExternalSource.source_id), unique=False)
    ext_constraint = PrimaryKeyConstraint("query_id", "source_id", name="source_key")
    score = Column(String, nullable=False)

    query_object = relationship(Query, uselist=False, backref=backref("external"), lazy="select")
    source_object = relationship(ExternalSource, uselist=False, backref=backref("external"), lazy="select")

    @hybrid_property
    def query(self):
        if self.query_object:
            return self.query_object.query_name
        else:
            return None

    @hybrid_property
    def valid_raw(self):
        if self.source_object:
            return self.source_object.valid_raw
        else:
            return None

    @hybrid_property
    def rtype(self):
        if self.source_object:
            return self.source_object.rtype
        else:
            return None

    @hybrid_property
    def source(self):
        if self.source_object:
            return self.source_object.source
        else:
            return None

    __table_args__ = ((ext_constraint,
                      Index("external_idx", "query_id", "source_id")))

    def __init__(self, query_id, source_id, score):

        self.query_id = query_id
        self.source_id = source_id
        if not isinstance(score, numbers.Number):
            raise sqlalchemy.exc.ArgumentError("Invalid score for external values: {}".format(type(score)))
        score = str(score)
        assert score.strip()
        self.score = str(score)


class ExternalSerializer:

    def __init__(self, handle,
                 configuration=None,
                 logger=None, delimiter="\t"):

        """
        :param handle: the file to be serialized.
        :type handle: str | io.IOBase | io.TextIOWrapper

        :param configuration: Optional configuration dictionary with db connection parameters.
        :type configuration: dict | MikadoConfiguration | DaijinConfiguration | None
        """

        if logger is not None:
            logger = check_logger(logger)
            self.logger = logger
        else:
            self.logger = create_default_logger("external")

        self.handle = handle

        if handle is None:
            self.logger.warning("No input file specified. Exiting.")
            self.close()
            return

        fasta_index = configuration.serialise.files.transcripts
        if isinstance(fasta_index, (str, bytes)):
            if isinstance(fasta_index, bytes):
                fasta_index = fasta_index.decode()
            if not os.path.exists(fasta_index):
                error = """I cannot find the mikado prepared FASTA file with the transcripts to analyse.
Please run mikado serialise in the folder with the correct files, and/or modify the configuration or the command line options."""
                self.logger.critical(error)
                raise AssertionError(error)
            self.fasta_index = pyfaidx.Fasta(fasta_index)
            # self.fasta_index = SeqIO.index(fasta_index, "fasta")
        elif fasta_index is None:
            self.logger.debug("No fasta index provided, we presume transcripts are already in the DB.")
            self.fasta_index = None
        elif isinstance(fasta_index, pyfaidx.Fasta):
            self.fasta_index = fasta_index
        else:
            error = "Unkwnown FASTA index: {}. I will presume that the transcripts are in the database".format(
                type(fasta_index))
            self.logger.warning(error)

        try:
            self.data = pd.read_csv(self.handle, delimiter=delimiter, index_col=["tid"])
        except (ValueError, pd.errors.ParserError, UnicodeDecodeError) as exc:
            self.logger.critical("Invalid input file!")
            self.logger.critical(exc)
            raise

        if len(self.data.columns) == 0:
            error = pd.errors.ParserError(
                "Not enough fields specified for the external file! Header: {}".format(self.data.columns))
            self.logger.exception(error)
            raise error

        self.data.fillna(0, inplace=True)

        self.engine = connect(configuration, logger=logger)

        session = Session(bind=self.engine, autocommit=False, autoflush=False, expire_on_commit=False)
        inspector = inspect(self.engine)
        if External.__tablename__ not in inspector.get_table_names():
            DBBASE.metadata.create_all(self.engine)  # @UndefinedVariable

        self.session = session
        self.maxobjects = configuration.serialise.max_objects

    def serialize(self):
        """This function will load the external scores into the database."""

        # First we have to load the sources inside the first table

        sources = dict()
        self.session.begin(subtransactions=True)

        # Check columns
        cols = []
        for col in self.data.columns:
            cols.append(re.sub(r"\.[0-9]*", '', str(col)))

        cols = Counter(cols)

        if cols.most_common()[0][1] > 1:  # IE the most common element is present more than one time
            raise IndexError("Duplicated values in the external table: {}".format(
                ",".join([_[0] for _ in cols.most_common() if _[1] > 1])
            ))

        for source in self.data.columns:

            if ((self.data[source].dtype == np.dtype("float") or
                    self.data[source].dtype == np.dtype("int")) and
                    0 <= self.data[source].min() <= self.data[source].max() <= 1):
                valid_raw = True
            else:
                valid_raw = False

            rtype = self.data[source].dtype
            if rtype == np.dtype("bool"):
                # We have to cast booleans as integers otherwise the conversion after extraction will fail
                self.data[source] = self.data[source].astype(int)

            source = ExternalSource(source, rtype=rtype, valid_raw=valid_raw)
            self.session.add(source)

        self.session.commit()

        # Now retrieve the values from the dictionary
        cache = dict()
        for source in self.session.query(ExternalSource):
            sources[source.source] = source.source_id
            continue
        for record in self.session.query(Query):
            cache[record.query_name] = record.query_id

        done = 0
        cache = self.load_fasta(cache)

        # Now we can serialise the content of the table
        objects = []
        decimal_place = 1
        for source in sources:
            for idx in self.data.index:
                objects.append(External(query_id=cache[idx],
                                        source_id=sources[source],
                                        score=self.data[source][idx]))
            if len(objects) >= self.maxobjects:
                done += len(objects)
                self.session.bulk_save_objects(objects)
                self.logger.debug("Serialised %d values", done)
                self.session.commit()
                objects = []
                if done >= self.data.shape[1] * self.data.shape[0] * 0.1 * decimal_place:
                    self.logger.info("Serialised %s%% of the external values",
                                     decimal_place * 10)
                    decimal_place += 1

        self.session.bulk_save_objects(objects, update_changed_only=False)
        done += len(objects)
        self.session.commit()
        self.logger.info("Finished serialising %d values", done)
        self.close()

    def __call__(self):
        """
        Alias for serialize()
        """

        self.serialize()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        Closing method for with. It will close the handles, if they are defined.
        :return:
        """

        if hasattr(self, "fasta_index") and self.fasta_index is not None:
            self.fasta_index.close()
        if hasattr(self, "session"):
            self.session.close()
        if hasattr(self, "engine"):
            self.engine.dispose()
        return

    def load_fasta(self, cache):
        """
        Private method to load data from the FASTA file into the database.
        :param cache: a dictionary which memoizes the IDs.
        :type cache: dict

        :return: the updated cache
        :rtype: dict

        """

        objects = []

        if self.fasta_index is not None:
            done = 0
            self.logger.info("%d entries already present in db, %d in the index",
                             len([fasta_key for fasta_key in self.fasta_index if
                                  fasta_key not in cache]),
                             len(self.fasta_index.keys()))
            found = set()
            for record in self.fasta_index.keys():
                if record in cache:
                    continue
                objects.append(Query(record, len(self.fasta_index[record])))
                assert record not in found, "The record {record} is duplicated in the FASTA index".format(record=record)
                found.add(record)
                if len(objects) >= self.maxobjects:
                    done += len(objects)
                    self.session.begin(subtransactions=True)
                    self.session.bulk_save_objects(objects)
                    self.session.commit()
                    self.logger.info(
                        "Loaded %d transcripts into query table", done)
                    objects = []

            done += len(objects)
            self.session.begin(subtransactions=True)
            self.session.bulk_save_objects(objects)
            self.session.commit()
            self.logger.info(
                "Finished loading %d transcripts into query table", done)

        # Now update cache
        cache.update(
            dict((_.query_name, _.query_id) for _ in iter(self.session.query(Query)))
        )

        assert len(cache) > 0

        return cache



#!/usr/bin/env python3
# coding: utf-8

"""Script to parse and retrieve given features from a GTF/GFF file."""

import sys
import argparse
import functools
import re


__author__ = 'Luca Venturini'


def print_gff_gene(curr_gene, curr_transcripts, args):
    """
    :param curr_gene: gene record
    :param curr_transcripts: dictionary of transcripts
    :param args: argparse
    :return: None
    """

    pat = re.compile("gene")

    if curr_gene is not None or len(curr_transcripts) > 0:
        starts, ends = [], []
        lines = []
        for tid in curr_transcripts:
            lines.append(str(curr_transcripts[tid][0]).rstrip())
            starts.append(curr_transcripts[tid][0].start)
            ends.append(curr_transcripts[tid][0].end)
            for rec in curr_transcripts[tid][1:]:
                rec.parent = [tid]
                lines.append(str(rec).rstrip())
        if starts and curr_gene is not None:
            curr_gene.start = min(starts)
        if ends and curr_gene is not None:
            curr_gene.end = max(ends)

        if curr_gene and not pat.search(curr_gene.feature) and not lines:
            print(curr_gene, file=args.out)
            print("###", file=args.out)
            return

        if lines:
            if curr_gene:
                print(curr_gene, file=args.out)
            print(*lines, sep="\n", file=args.out)
            print("###", file=args.out)


def verify_storability(record, mrna_ids, gene_ids, args):
    """
    This function verifies whether a GFF transcript has to be kept
    or not.
    :param record: the record to be evaluated
    :param gene_ids: a set of gene IDs
    :param mrna_ids: a set of mRNA IDs
    :param args: the namespace args
    :return:
    """

    bool_flag = False

    if args.reverse is False:
        if record.id in mrna_ids:
            bool_flag = True
        elif args.genes is True and any([p in gene_ids for p in record.parent]):
            bool_flag = True
            mrna_ids.add(record.id)
    elif args.reverse is True:
        if args.genes is False and record.id not in mrna_ids:
            bool_flag = True
        elif args.genes is True and not any([p in gene_ids for p in record.parent]):
            bool_flag = True
            mrna_ids.add(record.id)

    return bool_flag, mrna_ids


def grep_gff(args, gene_ids, mrna_ids):
    """
    Grep-like main function for *GFF3* files.
    :param args: argparse Namespace with the options.

    :param gene_ids: set of gene IDs to retain/discard.
    :type gene_ids: set

    :param mrna_ids: set of mRNA IDs to retain/discard.
    :type mrna_ids: set

    :return:
    """

    curr_gene = None
    curr_transcripts = dict()

    print("##gff-version 3", file=args.out)

    evaluator = functools.partial(verify_storability,
                                  **{"gene_ids": gene_ids,
                                     "args": args})

    for record in args.gff:
        if record.is_transcript is True:
            evaluated, mrna_ids = evaluator(record, mrna_ids)
            if evaluated is True:
                curr_transcripts[record.id] = [record]
        # Records with "derived from" are typically things like "protein" from TAIR
        elif record.is_exon is True or record.is_derived is True:
            if record.is_derived is True:
                parent_store = record.derived_from
            else:
                parent_store = record.parent
            for parent in parent_store:
                if parent in mrna_ids and args.reverse is False:
                    if parent not in curr_transcripts:
                        curr_transcripts[parent] = []
                    curr_transcripts[parent].append(record)
                elif parent not in mrna_ids and args.reverse is True:
                    if parent not in curr_transcripts:
                        curr_transcripts[parent] = []
                    curr_transcripts[parent].append(record)
        else:
            print_gff_gene(curr_gene, curr_transcripts, args)
            curr_gene = None
            curr_transcripts = dict()

            if record.id is None:
                continue
            curr_gene = record

    print_gff_gene(curr_gene, curr_transcripts, args)


def grep_gtf(args, gene_ids, mrna_ids):
    """
    Grep-like main function for *GTF* files.
    :param args:
    :return:

    :param gene_ids: set of gene IDs to retain/discard.
    :type gene_ids: set

    :param mrna_ids: set of mRNA IDs to retain/discard.
    :type mrna_ids: set

    """

    for record in args.gff:
        if not record:
            continue
        if args.genes is False:
            if record.transcript in mrna_ids and not args.reverse:
                print(record, file=args.out)
            elif record.transcript not in mrna_ids and args.reverse:
                print(record, file=args.out)
        else:
            if record.gene in gene_ids and not args.reverse:
                print(record, file=args.out)
            elif record.gene not in gene_ids and args.reverse:
                print(record, file=args.out)


def launch(args):
    """
    Function which chooses whether to use the grep_gtf
    or the grep_gff function for the analysis.
    :param args:
    :return:
    """
    from ...parsers import GFF, GTF, parser_factory

    args.gff = parser_factory(args.gff)
    gene_ids, mrna_ids = set(), set()
    for line in args.ids:
        if args.genes is False:
            mrna_id, gene_id = line.rstrip().split()[:2]
            mrna_ids.add(mrna_id)
            gene_ids.add(gene_id)
        else:
            gene_id = line.rstrip()
            gene_ids.add(gene_id)

    if isinstance(args.gff, GTF.GTF):
        grep_gtf(args, gene_ids, mrna_ids)
    elif isinstance(args.gff, GFF.GFF3):
        grep_gff(args, gene_ids, mrna_ids)
    else:
        raise TypeError(type(args.gff))


def grep_parser():
    """
    Simple command line parser for the utility.
    :return: args
    """

    parser = argparse.ArgumentParser(
        'Script to parse and retrieve given features from a GFF file.')
    parser.add_argument('-v', action='store_true', dest='reverse',
                        help="Exclude from the gff all the records in the id file.")
    parser.add_argument("--genes", action="store_true",
                        help="""Flag. If set, the program expects as ids
                        only a list of genes, and will exclude/include all the transcripts
                        children of the selected genes.""")
    parser.add_argument('ids', type=argparse.FileType(),
                        help="ID file (format: mrna_id, gene_id - tab separated)")
    parser.add_argument('gff', type=str, help="The GFF file to parse.")
    parser.add_argument('out', nargs='?',
                        type=argparse.FileType('w'), default=sys.stdout,
                        help="Optional output file")
    parser.set_defaults(func=launch)
    return parser



#!/usr/bin/env python3
# coding: utf-8

""" Script to calculate statistics about an annotation file.
It can take both GTF and GFF files as input."""

import sys
import argparse


__author__ = "Luca Venturini"


def launch(args):

    """
    Very simple launcher function, calls Calculator from this module.

    :param args: the argparse Namespace.
    """

    from ...scales.calculator import Calculator
    from ...parsers import parser_factory
    args.gff = parser_factory(args.gff)
    calculator = Calculator(args)
    calculator()


def stats_parser():

    """
    Argument parser.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('--only-coding', dest="only_coding", action="store_true", default=False)
    parser.add_argument('--tab-stats',
                        dest="tab_stats",
                        default=None,
                        type=argparse.FileType('w'),
                        help="Optional tabular file to write statistics for each transcript.")
    parser.add_argument("-v", "--verbose", action="store_true", default=False)
    parser.add_argument('gff', type=str, help="GFF file to parse.")
    parser.add_argument('out', type=argparse.FileType('w'), default=sys.stdout, nargs='?')
    parser.set_defaults(func=launch)
    return parser



#!/usr/bin/env python3
# coding: utf_8

"""
    This module defines the iterators that will parse BED12, GTF, GFF files.
"""

import io
import os
from ..exceptions import InvalidParsingFormat
from itertools import chain
from .parser import Parser
from . import GFF
from . import GTF
from . import bed12
from . import blast_utils
from . import bam_parser
from .bed12 import BED12


def parser_factory(string, input_format=None):
    """
    Function to recognize the input file type (GFF, GTF, BED12, BAM).
    :param string:
    :type string: (str|io.TextIOWrapper|io.BytesIO|io.BufferedReader|IO)
    :rtype: (GTF.GTF | GFF.GFF3 | bam_parser.BamParser | bed12.Bed12Parser)
    """

    streaming = False
    if isinstance(string, (io.BytesIO, io.BufferedReader, io.TextIOWrapper)):
        fname = "-"
        streaming = True
    elif isinstance(string, (bytes, str)):
        if isinstance(string, bytes):
            string = string.decode()
        fname = string
    else:
        raise ValueError("Invalid input type: {}".format(type(string)))

    order = [GTF.GTF, GFF.GFF3, bed12.Bed12Parser, bam_parser.BamParser]

    if input_format == "bam" or fname.endswith(".bam"):
        first = bam_parser.BamParser
        input_format = "bam"
    elif input_format == "gtf" or".gtf" in fname:
        first = GTF.GTF
        input_format = "gtf"
    elif input_format == "gff3" or ".gff" in fname or ".gff3" in fname:
        first = GFF.GFF3
        input_format = "gff3"
    elif input_format == "bed12" or ".bed12" in fname or ".bed" in fname:
        first = bed12.Bed12Parser
        input_format = "bed12"
    elif streaming is False:
        # We will have to impute.
        first = GTF.GTF
        input_format = "gtf"
    else:
        raise InvalidParsingFormat(
            "I cannot infer the correct file type from streaming. Please provide the correct file type.")

    if streaming is False and os.stat(fname).st_size == 0 or streaming is True:
        return first(fname)

    found = False
    raised = dict()
    for test in chain([first], [_ for _ in order if _ != first]):
        try:
            parser = test(fname)
            for row in parser:
                if test.__annot_type__ == "bam" or row.header is False:
                    found = True
                    break
                else:
                    continue
            if found:
                break
            else:
                raised[test.__annot_type__] = "No valid line found."
        except InvalidParsingFormat as exc:
            raised[test.__annot_type__] = exc
            continue

    if found and test.__annot_type__ == "bam":
        return test(fname)
    elif found:
        return test(string)
    else:

        raise InvalidParsingFormat(
            "Invalid file specified: {} should have been of format {}, but it could not be verified. Error:\n{}".format(
             fname if fname != "-" else "stream", input_format, raised[input_format]
            ))



"""
This module implements a very simple class for serialising the BLAST targets into
a table containing numeric ID, name and length.
"""


import collections
from sqlalchemy import Column, String, Integer
from .query import id_pattern
from ...utilities.dbutils import DBBASE
__author__ = 'Luca Venturini'


# These two classes are OK like this, they do not need more public methods!
# pylint: disable=too-few-public-methods
class Target(DBBASE):

    """
    Very simple serialization class for Target objects.
    """

    __tablename__ = "target"

    target_id = Column(Integer, primary_key=True)
    target_name = Column(String(200), unique=True, index=True)
    target_length = Column(Integer)
    named_tup = collections.namedtuple("Target",
                                       ["target_id", "target_name", "target_length"])

    def __init__(self, target_name, target_length):
        """
        Constructor method.
        :param target_name: name of the targets
        :type target_name: str

        :param target_length: length of the targets
        :type target_length: int
        """

        if not isinstance(target_name, str):
            raise TypeError("Invalid name: {0}".format(target_name))
        if not isinstance(target_length, int) or target_length <= 0:
            raise TypeError("Invalid length value: {0}".format(target_length))

        if id_pattern.search(target_name):
            target_name = id_pattern.search(target_name).groups()[0]

        self.target_name = target_name
        self.target_length = target_length

    def as_tuple(self):
        """Quick function to convert the SQLalchemy object
        into a named tuple with the same fields"""
        return self.named_tup(
            self.target_id, self.target_name, self.target_length)

    def as_dict(self):
        return {
            "target_name": self.target_name,
            "target_length": self.target_length
        }

# pylint: enable=too-few-public-methods



#!/usr/bin/env python3

import subprocess,sys,argparse,os
import shutil
import logging

def main():

    logging.basicConfig(format=" %(asctime)s - %(levelname)s - %(message)s",
                        level=logging.INFO)

    parser=argparse.ArgumentParser("Quick utility to rewrite the wrapper for CLASS.")
    parser.add_argument("--clean", default=False, action="store_true",
                        help="Flag. If set, remove tepmorary files.")
    parser.add_argument("--force", default=False, action="store_true",
                        help="Flag. If set, it forces recalculation of all intermediate files.")
    parser.add_argument("-c","--class_options", type=str, default='',
                        help="Additional options to be passed to CLASS. Default: no additional options.")
    parser.add_argument("-p", "--processors", type=int, default=1,
                        help="Number of processors to use with class.")
    parser.add_argument("--class_help", action="store_true",
                        default=False,
                        help="If called, the wrapper will ask class to display its help and exit.")
    parser.add_argument("-v", "--verbose", action="store_true", default=False)                        
    parser.add_argument("bam", type=argparse.FileType('rb'), default=None, nargs="?", help="Input BAM file.")
    parser.add_argument("out", nargs="?", type=argparse.FileType('wt'), default=sys.stdout,
                        help="Optional output file.")
    args=parser.parse_args()

    if args.class_help:
        print("Calling CLASS help..", file=sys.stderr)
        subprocess.call("class", shell=True)
        sys.exit(0)

    if args.bam is None:
        parser.error("The input BAM is required as an argument.")

    args.class_options+=" -p {0} ".format(args.processors) #Add the processors as argument
    logging.info("CLI: {0}".format(
        " ".join(sys.argv)))
    
    args.bam.close() #Quick and dirty check that the file exists.
    args.bam=os.path.abspath(args.bam.name) #Absolute position
    prefix=os.path.splitext(args.bam)[0] #Prefix without the .bam extension (comprehensive of dirs)

    #    if shutil.which("samtools") is None:
    # logging.debug("Loading the SAMTOOLS utility")
    # subprocess.call('source samtools-1.1', shell=True) #Load samtools if it is not present already
    #    else: pass
        
    # #    if shutil.which("junc") is None or shutil.which("class") is None:
    # logging.debug("Loading CLASS")
    # subprocess.call("source class-2.12", shell=True)
    # #   else: pass

    depth_file="{0}.depth".format(prefix)
    if not os.path.exists(depth_file) or args.force:
        if os.path.exists(depth_file):
            logging.warning("Deleting old depth file, because of --force option")
        with open(depth_file, 'wt') as depth_buffer:
            logging.info("Calculating depth with samtools")
            subprocess.call('samtools depth {0}'.format(args.bam), stdout=depth_buffer, shell=True)
    else:
        logging.warning("Depth file already present. Skipping this phase.")
    
    splice_file="{0}.splice".format(prefix)
    if not os.path.exists(splice_file) or args.force:
        if os.path.exists(splice_file):
            logging.warning("Deleting old splice file, because of --force option")
        with open(splice_file, 'wt') as splice_buffer:
            logging.info("Recovering junctions with the internal utility junc")
            subprocess.call("junc {0}".format(args.bam), stdout=splice_buffer, shell=True)
    else:
        logging.warning("Splice file already present. Skipping this phase.")


    logging.info("Launching the main process")
    class_cli="class {0} {1}".format(prefix, args.class_options)
    if args.verbose:
        class_cli+=" --verbose "
    logging.info("CLASS CLI:\n\t{0}".format(class_cli))
    class_sub=subprocess.Popen(class_cli, shell=True,
                    stdout=args.out, stderr=subprocess.PIPE)

    for line in class_sub.stderr:
        line=line.decode().rstrip()
        logging.info("CLASS message:\n{0}".format(line))

    if args.clean:
        logging.info("Removing temporary files..")
        os.remove(depth_file)
        logging.debug("Removed the DEPTH file")
        os.remove(splice_file)
        logging.debug("Removed the SPLICE file")

    logging.info("CLASS finished!")
    return

if __name__=='__main__': main()



"""
TestCase to test the DButils module
"""

import os
import unittest

from sqlalchemy import inspect
from sqlalchemy.engine import Inspector

from ..utilities import dbutils
from ..configuration import configurator
import sqlalchemy
import sqlalchemy.orm.session
import sqlalchemy.engine.base
import sqlite3
from .. import serializers
import shutil
import pkg_resources


__author__ = 'Luca Venturini'


class TestDbConnect(unittest.TestCase):

    def setUp(self):
        self.json = configurator.load_and_validate_config(
            os.path.join(os.path.dirname(__file__), "configuration.yaml"))
        self.json.db_settings.db = pkg_resources.resource_filename("Mikado.tests", "mikado.db")

    def test_connector(self):
        connector = dbutils.create_connector(self.json)
        self.assertIsInstance(connector, sqlite3.Connection)

    def test_engine(self):

        engine = dbutils.connect(self.json)
        self.assertIsInstance(engine, sqlalchemy.engine.base.Engine)
        table_names = ['chrom', 'hit', "orf", "external",
                       "external_sources", 'hsp', 'junctions', 'query', 'target']
        self.assertEqual(sorted(inspect(engine).get_table_names()), sorted(table_names))

    def test_content(self):

        engine = dbutils.connect(self.json)
        sessionmaker = sqlalchemy.orm.sessionmaker(bind=engine)
        session = sessionmaker()
        # Simple tests based on the static content of the dictionary
        self.assertEqual(session.query(serializers.junction.Junction).count(), 371,
                         self.json.db_settings)
        self.assertEqual(session.query(serializers.orf.Orf).count(), 268)
        self.assertEqual(session.query(serializers.blast_serializer.Target).count(), 42992)
        self.assertEqual(session.query(serializers.blast_serializer.Query).count(), 112)
        self.assertEqual(session.query(serializers.blast_serializer.Hit).count(), 674)
        self.assertEqual(session.query(serializers.blast_serializer.Hsp).count(), 674)

        first_query = session.query(serializers.blast_serializer.Query).limit(1).one()
        astup = first_query.as_tuple()
        self.assertTrue(astup._fields, ("query_id", "query_name", "query_length"))
        self.assertIsInstance(astup.query_id, int)
        self.assertIsInstance(astup.query_length, int)
        self.assertIsInstance(astup.query_name, str)
        
        first_target = session.query(
            serializers.blast_serializer.Target).limit(1).one()
        astup = first_target.as_tuple()
        self.assertTrue(astup._fields, ("target_id", "target_name", "target_length"))
        self.assertIsInstance(astup.target_id, int)
        self.assertIsInstance(astup.target_length, int)
        self.assertIsInstance(astup.target_name, str)

    def test_query_init(self):

        _ = serializers.blast_serializer.Query("foo", 1000)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Query(100, 1000)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Query("foo", 0)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Query("foo", -10)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Query("foo", 1000.0)

    def test_target_init(self):

        _ = serializers.blast_serializer.Target("foo", 1000)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Target(100, 1000)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Target("foo", 0)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Target("foo", -10)
        with self.assertRaises(TypeError):
            _ = serializers.blast_serializer.Target("foo", 1000.0)

    def test_wrong_db(self):
        self.json.db_settings.dbtype = "sqlite_foo"
        with self.assertRaises(ValueError):
            _ = dbutils.create_connector(self.json)

    @unittest.skipUnless(os.path.exists("/dev/shm"),
                         "/dev/shm is not available on this system.")
    def test_connect_to_shm(self):
        self.json.pick.run_options.shm = True
        shutil.copy(self.json.db_settings.db, "/dev/shm/")
        self.json.db_settings.db = os.path.join("/dev/shm/",
                                                      os.path.basename(self.json.db_settings.db))
        connector = dbutils.connect(self.json)
        self.assertEqual(str(connector.url), "sqlite://")
        engine = dbutils.connect(self.json)
        sessionmaker = sqlalchemy.orm.sessionmaker(bind=engine)
        session = sessionmaker()
        first_target = session.query(serializers.blast_serializer.Target).limit(1).one()
        astup = first_target.as_tuple()
        self.assertTrue(astup._fields, ("target_id", "target_name", "target_length"))
        os.remove(os.path.join("/dev/shm/", os.path.basename(self.json.db_settings.db)))

    def test_to_memory(self):
        engine = dbutils.connect(None)
        self.assertEqual(str(engine.url), "sqlite:///:memory:")


if __name__ == "__main__":
    unittest.main()
